# Copyright (C) 2011-2012, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.document import Drawing
from ezdxf.addons.tablepainter import (
    TablePainter,
    CustomCell,
    Grid,
    CellStyle,
    DEFAULT_BORDER_COLOR,
)


@pytest.fixture(scope="module")
def doc() -> Drawing:
    return ezdxf.new("R12")


class MockCell(CustomCell):
    counter = 0

    def render(self, layout, coords, layer):
        MockCell.counter += 1

    @staticmethod
    def reset():
        MockCell.counter = 0


def test_init():
    table = TablePainter((0, 0), 10, 10, default_grid=False)
    assert bool(table) is True
    style = table.get_cell_style("default")

    assert style.left.status is False
    assert style.right.status is False
    assert style.top.status is False
    assert style.bottom.status is False

    table = TablePainter((0, 0), 10, 10, default_grid=True)
    style = table.get_cell_style("default")
    assert style.left.status is True
    assert style.right.status is True
    assert style.top.status is True
    assert style.bottom.status is True


def test_setter_methods():
    table = TablePainter((0, 0), 10, 10)
    table.set_col_width(0, 3.0)
    assert table.col_widths[0] == 3.0
    table.set_row_height(0, 4.0)
    assert table.row_heights[0] == 4.0


def test_cell_index():
    table = TablePainter((0, 0), 10, 10)
    with pytest.raises(IndexError):
        table.get_cell(10, 10)
    with pytest.raises(IndexError):
        table.get_cell(-1, 10)


def test_default_text_cell():
    table = TablePainter((0, 0), 10, 10)
    text_cell = table.text_cell(0, 0, "test")
    assert text_cell.text == "test"
    cell = table.get_cell(0, 0)
    assert cell.span == (1, 1)
    assert cell.stylename == "default"


def test_text_cell():
    table = TablePainter((0, 0), 10, 10)
    text_cell = table.text_cell(8, 8, "test88", span=(2, 2), style="extrastyle")
    assert text_cell.text == "test88"
    cell = table.get_cell(8, 8)
    assert cell.span == (2, 2)
    assert cell.stylename == "extrastyle"


def test_block_cell(doc):
    block = doc.blocks.new("EMPTY_BLOCK")
    table = TablePainter((0, 0), 10, 10)
    block_cell = table.block_cell(1, 1, block, span=(3, 3))
    assert block_cell.block_name == "EMPTY_BLOCK"

    cell = table.get_cell(1, 1)
    assert cell.span == (3, 3)
    assert cell.stylename == "default"


def test_frame():
    table = TablePainter((0, 0), 10, 10)
    frame = table.frame(0, 0, width=10, height=2)
    assert frame.pos == (0, 0)
    assert frame.span == (2, 10)


def test_cell_style():
    table = TablePainter((0, 0), 10, 10)
    table.new_cell_style("extra", text_color=199)
    style = table.get_cell_style("extra")
    assert style.text_color == 199
    with pytest.raises(KeyError):
        table.get_cell_style("extraextra")


def test_border_style():
    table = TablePainter((0, 0), 10, 10)
    border_style = table.new_border_style(
        color=1, status=True, linetype="DOT", priority=99
    )
    assert border_style.color == 1
    assert border_style.status is True
    assert border_style.linetype == "DOT"
    assert border_style.priority == 99


def test_visibility_map():
    from ezdxf.addons.tablepainter import VisibilityMap

    table = TablePainter((0, 0), 3, 3)
    text_cell = table.text_cell(0, 0, "text", span=(2, 2))
    vmap = VisibilityMap(table)
    empty = table.empty_cell
    expected = [
        (0, 0, text_cell),
        (0, 2, empty),  # cell (0, 1) is covered by (0,0)
        (1, 2, empty),  # cells (1, 0), (1, 2) are covered by cell (0, 0)
        (2, 0, empty),
        (2, 1, empty),
        (2, 2, empty),
    ]  # row 2
    for got, should in zip(table.iter_visible_cells(vmap), expected):
        assert got[0] == should[0]  # row
        assert got[1] == should[1]  # col
        assert got[2] == should[2]  # cell


def test_rendering(doc):
    MockCell.reset()
    layout = doc.blocks.new("test_rendering")
    table = TablePainter((0, 0), 3, 3)
    indices = [
        (0, 0),
        (0, 1),
        (0, 2),
        (1, 0),
        (1, 1),
        (1, 2),
        (2, 0),
        (2, 1),
        (2, 2),
    ]
    cell = MockCell(table, "default", (1, 1))
    for row, col in indices:
        table.set_cell(row, col, cell)
    table.render(layout)
    assert cell.counter == 9  # count get_dxf_entity calls


def test_dxf_creation_span(doc):
    MockCell.reset()
    layout = doc.blocks.new("test_dxf_creation_span")
    table = TablePainter((0, 0), 3, 3)
    indices = [
        (0, 0),
        (0, 1),
        (0, 2),
        (1, 0),
        (1, 1),
        (1, 2),
        (2, 0),
        (2, 1),
        (2, 2),
    ]
    cell = MockCell(table, "default", (1, 1))
    for row, col in indices:
        table.set_cell(row, col, cell)
    spancell = MockCell(table, "default", span=(2, 2))  # hides 3 cells
    table.set_cell(0, 0, spancell)
    table.render(layout)
    assert cell.counter == 6  # count get_dxf_entity calls


def test_span_beyond_table_borders(doc):
    layout = doc.blocks.new("test_span_beyond_table_borders")
    table = TablePainter((0, 0), 3, 3)
    table.text_cell(0, 2, "ERROR", span=(1, 2))
    with pytest.raises(IndexError):
        table.render(layout)
    table.text_cell(2, 0, "ERROR", span=(2, 1))
    with pytest.raises(IndexError):
        table.render(layout)


@pytest.fixture
def table():
    table = TablePainter((0, 0), 3, 3)
    for x in range(3):
        table.set_col_width(x, 3.0)
        table.set_row_height(x, 3.0)
    return table


def test_grid_coords(table):
    grid = Grid(table)
    left, right, top, bottom = grid.cell_coords(1, 1, span=(1, 1))
    assert left == 3.0  # , places=4)
    assert right == 6.0
    assert top == -3.0
    assert bottom == -6.0


def test_grid_coords_span(table):
    grid = Grid(table)
    left, right, top, bottom = grid.cell_coords(0, 0, span=(2, 2))
    assert left == 0.0
    assert right == 6.0
    assert top == 0.0
    assert bottom == -6.0


def test_draw_cell_background(doc, table):
    grid = Grid(table)
    layout = doc.blocks.new("test_draw_cell_background")
    table.new_cell_style("fill", bg_color=17)
    cell = table.get_cell(0, 0)
    cell.stylename = "fill"
    grid.render_cell_background(layout, 0, 0, cell)
    solid = list(layout)[0]
    assert solid.dxftype() == "SOLID"


def test_set_border_status():
    style = CellStyle()
    style.set_border_status(False, True, False, True)
    assert style.left.status is False
    assert style.right.status is True
    assert style.top.status is False
    assert style.bottom.status is True


def test_set_border_style():
    style = CellStyle()
    border_style = CellStyle.get_default_border_style()
    border_style.color = 17

    style.set_border_style(border_style, False, True, False, True)
    assert style.left.color == DEFAULT_BORDER_COLOR
    assert style.right.color == 17
    assert style.top.color == DEFAULT_BORDER_COLOR
    assert style.bottom.color == 17


class TestStyle:
    def test_init_default_style(self):
        style = CellStyle()
        assert style.text_style == "STANDARD"

    def test_init_custom_style(self):
        style = CellStyle({"text_style": "Arial"})
        assert style.text_style == "Arial"

    def test_get_attribute_index_operator(self):
        style = CellStyle()
        assert style["text_style"] == style.text_style

    def test_invalid_attribute_name_raises_key_error(self):
        style = CellStyle()
        with pytest.raises(KeyError):  # get
            _ = style["xxx"]

        with pytest.raises(KeyError):  # set
            style["xxx"] = "xxx"



# Copyright (c) 2020-2023, Manfred Moitzi
# License: MIT License
# type: ignore
from __future__ import annotations
from typing import Optional
import logging
import os
import platform
import shutil
import subprocess
import tempfile
import time
from contextlib import contextmanager
from pathlib import Path

import ezdxf
from ezdxf.document import Drawing
from ezdxf.lldxf.validator import (
    is_dxf_file,
    dxf_info,
    is_binary_dxf_file,
    dwg_version,
)

logger = logging.getLogger("ezdxf")


class ODAFCError(IOError):
    pass


class UnknownODAFCError(ODAFCError):
    pass


class ODAFCNotInstalledError(ODAFCError):
    pass


class UnsupportedFileFormat(ODAFCError):
    pass


class UnsupportedPlatform(ODAFCError):
    pass


class UnsupportedVersion(ODAFCError):
    pass


VERSION_MAP = {
    "R12": "ACAD12",
    "R13": "ACAD13",
    "R14": "ACAD14",
    "R2000": "ACAD2000",
    "R2004": "ACAD2004",
    "R2007": "ACAD2007",
    "R2010": "ACAD2010",
    "R2013": "ACAD2013",
    "R2018": "ACAD2018",
    "AC1004": "ACAD9",
    "AC1006": "ACAD10",
    "AC1009": "ACAD12",
    "AC1012": "ACAD13",
    "AC1014": "ACAD14",
    "AC1015": "ACAD2000",
    "AC1018": "ACAD2004",
    "AC1021": "ACAD2007",
    "AC1024": "ACAD2010",
    "AC1027": "ACAD2013",
    "AC1032": "ACAD2018",
}

VALID_VERSIONS = {
    "ACAD9",
    "ACAD10",
    "ACAD12",
    "ACAD13",
    "ACAD14",
    "ACAD2000",
    "ACAD2004",
    "ACAD2007",
    "ACAD2010",
    "ACAD2013",
    "ACAD2018",
}

WINDOWS = "Windows"
LINUX = "Linux"
DARWIN = "Darwin"


def get_win_exec_path() -> str:
    return ezdxf.options.get("odafc-addon", "win_exec_path").strip('"')


def get_unix_exec_path() -> str:
    return ezdxf.options.get("odafc-addon", "unix_exec_path").strip('"')


def is_installed() -> bool:
    """Returns ``True`` if the ODAFileConverter is installed."""
    if platform.system() in (LINUX, DARWIN):
        unix_exec_path = get_unix_exec_path()
        if unix_exec_path and Path(unix_exec_path).is_file():
            return True
        return shutil.which("ODAFileConverter") is not None
    # Windows:
    return os.path.exists(get_win_exec_path())


def map_version(version: str) -> str:
    return VERSION_MAP.get(version.upper(), version.upper())


def readfile(
    filename: str | os.PathLike, version: Optional[str] = None, *, audit: bool = False
) -> Drawing:
    """Uses an installed `ODA File Converter`_ to convert a DWG/DXB/DXF file
    into a temporary DXF file and load this file by `ezdxf`.

    Args:
        filename: file to load by ODA File Converter
        version: load file as specific DXF version, by default the same version
            as the source file or if not detectable the latest by `ezdxf`
            supported version.
        audit: audit source file before loading

    Raises:
        FileNotFoundError: source file not found
        odafc.UnknownODAFCError: conversion failed for unknown reasons
        odafc.UnsupportedVersion: invalid DWG version specified
        odafc.UnsupportedFileFormat: unsupported file extension
        odafc.ODAFCNotInstalledError: ODA File Converter not installed

    """
    infile = Path(filename).absolute()
    if not infile.is_file():
        raise FileNotFoundError(f"No such file: '{infile}'")
    if isinstance(version, str):
        version = map_version(version)
    else:
        version = _detect_version(filename)

    with tempfile.TemporaryDirectory(prefix="odafc_") as tmp_dir:
        args = _odafc_arguments(
            infile.name,
            str(infile.parent),
            tmp_dir,
            output_format="DXF",
            version=version,
            audit=audit,
        )
        _execute_odafc(args)
        out_file = Path(tmp_dir) / infile.with_suffix(".dxf").name
        if out_file.exists():
            doc = ezdxf.readfile(str(out_file))
            doc.filename = str(infile.with_suffix(".dxf"))
            return doc
    raise UnknownODAFCError("Failed to convert file: Unknown Error")


def export_dwg(
    doc: Drawing,
    filename: str | os.PathLike,
    version: Optional[str] = None,
    *,
    audit: bool = False,
    replace: bool = False,
) -> None:
    """Uses an installed `ODA File Converter`_ to export the DXF document `doc`
    as a DWG file.

    A temporary DXF file will be created and converted to DWG by the
    ODA File Converter. If `version` is not specified the DXF version of the
    source document is used.

    Args:
        doc: `ezdxf` DXF document as :class:`~ezdxf.drawing.Drawing` object
        filename: output DWG filename, the extension will be set to ".dwg"
        version: DWG version to export, by default the same version as
            the source document.
        audit: audit source file by ODA File Converter at exporting
        replace: replace existing DWG file if ``True``

    Raises:
        FileExistsError: target file already exists, and argument `replace` is
            ``False``
        FileNotFoundError: parent directory of target file does not exist
        odafc.UnknownODAFCError: exporting DWG failed for unknown reasons
        odafc.ODAFCNotInstalledError: ODA File Converter not installed

    """
    if version is None:
        version = doc.dxfversion
    export_version = VERSION_MAP[version]
    dwg_file = Path(filename).absolute()
    out_folder = Path(dwg_file.parent)
    if dwg_file.exists():
        if replace:
            dwg_file.unlink()
        else:
            raise FileExistsError(f"File already exists: {dwg_file}")
    if out_folder.exists():
        with tempfile.TemporaryDirectory(prefix="odafc_") as tmp_dir:
            dxf_file = Path(tmp_dir) / dwg_file.with_suffix(".dxf").name

            # Save DXF document
            old_filename = doc.filename
            doc.saveas(dxf_file)
            doc.filename = old_filename

            arguments = _odafc_arguments(
                dxf_file.name,
                tmp_dir,
                str(out_folder),
                output_format="DWG",
                version=export_version,
                audit=audit,
            )
            _execute_odafc(arguments)
    else:
        raise FileNotFoundError(f"No such file or directory: '{str(out_folder)}'")


def convert(
    source: str | os.PathLike,
    dest: str | os.PathLike = "",
    *,
    version="R2018",
    audit=True,
    replace=False,
):
    """Convert `source` file to `dest` file.

    The file extension defines the target format
    e.g. :code:`convert("test.dxf", "Test.dwg")` converts the source file to a
    DWG file.
    If `dest` is an empty string the conversion depends on the source file
    format and is DXF to DWG or DWG to DXF.
    To convert DXF to DXF an explicit destination filename is required:
    :code:`convert("r12.dxf", "r2013.dxf", version="R2013")`

    Args:
        source: source file
        dest: destination file, an empty string uses the source filename with
            the extension of the target format e.g. "test.dxf" -> "test.dwg"
        version: output DXF/DWG version e.g. "ACAD2018", "R2018", "AC1032"
        audit: audit files
        replace: replace existing destination file

    Raises:
        FileNotFoundError: source file or destination folder does not exist
        FileExistsError: destination file already exists and argument `replace`
            is ``False``
        odafc.UnsupportedVersion: invalid DXF version specified
        odafc.UnsupportedFileFormat: unsupported file extension
        odafc.UnknownODAFCError: conversion failed for unknown reasons
        odafc.ODAFCNotInstalledError: ODA File Converter not installed

    """
    version = map_version(version)
    if version not in VALID_VERSIONS:
        raise UnsupportedVersion(f"Invalid version: '{version}'")
    src_path = Path(source).expanduser().absolute()
    if not src_path.exists():
        raise FileNotFoundError(f"Source file not found: '{source}'")
    if dest:
        dest_path = Path(dest)
    else:
        ext = src_path.suffix.lower()
        if ext == ".dwg":
            dest_path = src_path.with_suffix(".dxf")
        elif ext == ".dxf":
            dest_path = src_path.with_suffix(".dwg")
        else:
            raise UnsupportedFileFormat(f"Unsupported file format: '{ext}'")
    if dest_path.exists() and not replace:
        raise FileExistsError(f"Target file already exists: '{dest_path}'")
    parent_dir = dest_path.parent
    if not parent_dir.exists() or not parent_dir.is_dir():
        # Cannot copy result to destination folder!
        raise FileNotFoundError(f"Destination folder does not exist: '{parent_dir}'")
    ext = dest_path.suffix
    fmt = ext.upper()[1:]
    if fmt not in ("DXF", "DWG"):
        raise UnsupportedFileFormat(f"Unsupported file format: '{ext}'")

    with tempfile.TemporaryDirectory(prefix="odafc_") as tmp_dir:
        arguments = _odafc_arguments(
            src_path.name,
            in_folder=str(src_path.parent),
            out_folder=str(tmp_dir),
            output_format=fmt,
            version=version,
            audit=audit,
        )
        _execute_odafc(arguments)
        result = list(Path(tmp_dir).iterdir())
        if result:
            try:
                shutil.move(result[0], dest_path)
            except IOError:
                shutil.copy(result[0], dest_path)
        else:
            UnknownODAFCError(f"Unknown error: no {fmt} file was created")


def _detect_version(path: str) -> str:
    """Returns the DXF/DWG version of file `path` as ODAFC compatible version
    string.

    Raises:
        odafc.UnsupportedVersion: unknown or unsupported DWG version
        odafc.UnsupportedFileFormat; unsupported file extension

    """
    version = "ACAD2018"
    ext = os.path.splitext(path)[1].lower()
    if ext == ".dxf":
        if is_binary_dxf_file(path):
            pass
        elif is_dxf_file(path):
            with open(path, "rt") as fp:
                info = dxf_info(fp)
                version = VERSION_MAP[info.version]
    elif ext == ".dwg":
        version = dwg_version(path)
        if version is None:
            raise UnsupportedVersion("Unknown or unsupported DWG version.")
    else:
        raise UnsupportedFileFormat(f"Unsupported file format: '{ext}'")

    return map_version(version)


def _odafc_arguments(
    filename: str,
    in_folder: str,
    out_folder: str,
    output_format: str = "DXF",
    version: str = "ACAD2013",
    audit: bool = False,
) -> list[str]:
    """ODA File Converter command line format:

    ODAFileConverter "Input Folder" "Output Folder" version type recurse audit [filter]

        - version: output version: "ACAD9" - "ACAD2018"
        - type: output file type: "DWG", "DXF", "DXB"
        - recurse: recurse Input Folder: "0" or "1"
        - audit: audit each file: "0" or "1"
        - optional Input files filter: default "*.DWG,*.DXF"

    """
    recurse = "0"
    audit_str = "1" if audit else "0"
    return [
        in_folder,
        out_folder,
        version,
        output_format,
        recurse,
        audit_str,
        filename,
    ]


def _get_odafc_path(system: str) -> str:
    """Get ODAFC application path.

    Raises:
        odafc.ODAFCNotInstalledError: ODA File Converter not installed

    """
    # on Linux and Darwin check if UNIX_EXEC_PATH is set and exist and
    # return this path as exec path.
    # This may help if the "which" command can not find the "ODAFileConverter"
    # command and also adds support for AppImages provided by ODA.
    unix_exec_path = get_unix_exec_path()

    if system != WINDOWS and unix_exec_path:
        if Path(unix_exec_path).is_file():
            return unix_exec_path
        else:
            logger.warning(
                f"command '{unix_exec_path}' not found, using 'ODAFileConverter'"
            )

    path = shutil.which("ODAFileConverter")
    if not path and system == WINDOWS:
        path = get_win_exec_path()
        if not Path(path).is_file():
            path = None

    if not path:
        raise ODAFCNotInstalledError(
            f"Could not find ODAFileConverter in the path. "
            f"Install application from https://www.opendesign.com/guestfiles/oda_file_converter"
        )
    return path


@contextmanager
def _linux_dummy_display():
    """See xvbfwrapper library for a more feature complete xvfb interface."""
    if shutil.which("Xvfb"):
        display = f":{os.getpid()}"  # Each process has its own screen id
        proc = subprocess.Popen(
            ["Xvfb", display, "-screen", "0", "800x600x24"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        time.sleep(0.1)
        yield display
        try:
            proc.terminate()
            proc.wait()
        except OSError:
            pass
    else:
        logger.warning(f"Install xvfb to prevent the ODAFileConverter GUI from opening")
        yield os.environ["DISPLAY"]


def _run_with_no_gui(
    system: str, command: str, arguments: list[str]
) -> subprocess.Popen:
    """Execute ODAFC application without launching the GUI.

    Args:
        system: "Linux", "Windows" or "Darwin"
        command: application to execute
        arguments: ODAFC argument list

    Raises:
        odafc.UnsupportedPlatform: for unsupported platforms
        odafc.ODAFCNotInstalledError: ODA File Converter not installed

    """
    if system == LINUX:
        with _linux_dummy_display() as display:
            env = os.environ.copy()
            env["DISPLAY"] = display
            proc = subprocess.run(
                [command] + arguments, text=True, capture_output=True, env=env
            )

    elif system == DARWIN:
        # TODO: unknown how to prevent the GUI from appearing on macOS
        proc = subprocess.run([command] + arguments, text=True, capture_output=True)

    elif system == WINDOWS:
        # New code from George-Jiang to solve the GUI pop-up problem
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags = (
            subprocess.CREATE_NEW_CONSOLE | subprocess.STARTF_USESHOWWINDOW
        )
        startupinfo.wShowWindow = subprocess.SW_HIDE
        proc = subprocess.Popen(
            [command] + arguments,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            startupinfo=startupinfo,
        )
        proc.wait()
    else:
        # The OpenDesign Alliance only provides executables for Linux, macOS
        # and Windows:
        raise UnsupportedPlatform(f"Unsupported platform: {system}")
    return proc


def _odafc_failed(system: str, returncode: int, stderr: str) -> bool:
    # changed v0.18.1, see https://github.com/mozman/ezdxf/issues/707
    stderr = stderr.strip()
    if system == LINUX:
        # ODAFileConverter *always* crashes on Linux even if the output was successful
        return stderr != "" and stderr != "Quit (core dumped)"
    elif returncode != 0:
        return True
    else:
        return stderr != ""


def _execute_odafc(arguments: list[str]) -> Optional[bytes]:
    """Execute ODAFC application.

    Args:
        arguments: ODAFC argument list

    Raises:
        odafc.ODAFCNotInstalledError: ODA File Converter not installed
        odafc.UnknownODAFCError: execution failed for unknown reasons
        odafc.UnsupportedPlatform: for unsupported platforms

    """
    logger.debug(f"Running ODAFileConverter with arguments: {arguments}")
    system = platform.system()
    oda_fc = _get_odafc_path(system)
    proc = _run_with_no_gui(system, oda_fc, arguments)
    returncode = proc.returncode
    if system == WINDOWS:
        stdout = proc.stdout.read().decode("utf-8")
        stderr = proc.stderr.read().decode("utf-8")
    else:
        stdout = proc.stdout
        stderr = proc.stderr

    if _odafc_failed(system, returncode, stderr):
        msg = (
            f"ODA File Converter failed: return code = {returncode}.\n"
            f"stdout: {stdout}\nstderr: {stderr}"
        )
        logger.debug(msg)
        raise UnknownODAFCError(msg)
    return stdout



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
"""xplayer = cross backend player."""
from __future__ import annotations
from typing import Callable
from ezdxf.math import Vec2
from ezdxf.colors import RGB

from ezdxf.addons.drawing.backend import BackendInterface, BkPath2d
from ezdxf.addons.drawing.properties import BackendProperties
from ezdxf.addons.hpgl2 import api as hpgl2
from ezdxf.addons.hpgl2.backend import (
    Properties as HPGL2Properties,
    RecordType as HPGL2RecordType,
)


def hpgl2_to_drawing(
    player: hpgl2.Player,
    backend: BackendInterface,
    bg_color: str = "#ffffff",
    override: Callable[[BackendProperties], BackendProperties] | None = None,
) -> None:
    """Replays the recordings of the HPGL2 Recorder on a backend of the drawing add-on."""
    if bg_color:
        backend.set_background(bg_color)
    for record_type, properties, record_data in player.recordings():
        backend_properties = _make_drawing_backend_properties(properties)
        if override:
            backend_properties = override(backend_properties)
        if record_type == HPGL2RecordType.POLYLINE:
            points: list[Vec2] = record_data.vertices()
            size = len(points)
            if size == 1:
                backend.draw_point(points[0], backend_properties)
            elif size == 2:
                backend.draw_line(points[0], points[1], backend_properties)
            else:
                backend.draw_path(BkPath2d.from_vertices(points), backend_properties)
        elif record_type == HPGL2RecordType.FILLED_PATHS:
            backend.draw_filled_paths(record_data, backend_properties)
        elif record_type == HPGL2RecordType.OUTLINE_PATHS:
            for p in record_data:
                backend.draw_path(p, backend_properties)
    backend.finalize()


def _make_drawing_backend_properties(properties: HPGL2Properties) -> BackendProperties:
    """Make BackendProperties() for the drawing add-on."""
    return BackendProperties(
        color=properties.pen_color.to_hex(),
        lineweight=properties.pen_width,
        layer="0",
        pen=properties.pen_index,
        handle="",
    )


def map_color(color: str) -> Callable[[BackendProperties], BackendProperties]:
    def _map_color(properties: BackendProperties) -> BackendProperties:
        return BackendProperties(
            color=color,
            lineweight=properties.lineweight,
            layer=properties.layer,
            pen=properties.pen,
            handle=properties.handle,
        )

    return _map_color


def map_monochrome(dark_mode=True) -> Callable[[BackendProperties], BackendProperties]:
    def to_gray(color: str) -> str:
        gray = round(RGB.from_hex(color).luminance * 255)
        if dark_mode:
            gray = 255 - gray
        return RGB(gray, gray, gray).to_hex()

    def _map_color(properties: BackendProperties) -> BackendProperties:
        color = properties.color
        alpha = color[7:9]
        return BackendProperties(
            color=to_gray(color[:7]) + alpha,
            lineweight=properties.lineweight,
            layer=properties.layer,
            pen=properties.pen,
            handle=properties.handle,
        )

    return _map_color



# Created: 16.03.2011, 2018 rewritten for pytest
# Copyright (C) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.entities.vport import VPort


@pytest.fixture
def vport():
    return VPort.new(
        "FFFF",
        dxfattribs={
            "name": "VP1",
        },
    )


def test_name(vport):
    assert vport.dxf.name == "VP1"



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License

from ezdxf.tools import suppress_zeros


def test_leading_zeros():
    assert suppress_zeros("0") == "0"
    assert suppress_zeros("0.", leading=True, trailing=True) == "0"
    assert suppress_zeros("0.0", leading=True, trailing=True) == "0"
    assert suppress_zeros(".0", leading=True, trailing=True) == "0"
    assert suppress_zeros("-.0", leading=True, trailing=True) == "0"
    assert suppress_zeros("+.0", leading=True, trailing=True) == "0"

    assert suppress_zeros("0.0", leading=False, trailing=True) == "0"
    assert suppress_zeros("0.1", leading=False, trailing=True) == "0.1"
    assert suppress_zeros("1.0", leading=False, trailing=True) == "1"
    assert suppress_zeros("1.0000", leading=False, trailing=True) == "1"
    assert suppress_zeros("-1.", leading=False, trailing=True) == "-1"

    assert suppress_zeros("0.10", leading=True, trailing=False) == ".10"
    assert suppress_zeros("-0.10", leading=True, trailing=False) == "-.10"
    assert suppress_zeros("+000.10", leading=True, trailing=False) == "+.10"


def test_trim_trailing_zeros_for_integers():
    assert suppress_zeros("0", trailing=True) == "0"
    assert suppress_zeros("1", trailing=True) == "1"
    assert suppress_zeros("-1", trailing=True) == "-1"
    assert suppress_zeros("10", trailing=True) == "10"
    assert suppress_zeros("-10", trailing=True) == "-10"



# This test is hard to do in pytest!

from pathlib import Path
import ezdxf

p = Path(__file__).with_name("disable.ini")

ezdxf.options.read_file(str(p))
print(f"disable C-Extension (should be True): {ezdxf.options.disable_c_ext}")
assert ezdxf.options.disable_c_ext is True

# It is not possible to deactivate the C-extension by a user config file
# loaded after the ezdxf import, because the setup process of ezdxf is already
# finished.
print(f"using C-Extension (should be True): {ezdxf.options.use_c_ext}")
assert ezdxf.options.use_c_ext is True



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Sequence, Iterator
import math

from .deps import (
    Vec2,
    Path,
    NULLVEC2,
    ConstructionCircle,
    Bezier4P,
)
from .properties import RGB, Properties, FillType
from .backend import Backend
from .polygon_buffer import PolygonBuffer
from .page import Page


class Plotter:
    """
    The :class:`Plotter` class represents a virtual plotter device.

    The HPGL/2 commands send by the :class:`Interpreter` are processed into simple
    polylines and filled polygons and send to low level :class:`Backend`.

    HPGL/2 uses a units system called "Plot Units":

    - 1 plot unit (plu) = 0.025mm
    - 40 plu = 1 mm
    - 1016 plu = 1 inch

    The Plotter device does not support font rendering and page rotation (RO).
    The scaling commands IP, RP, SC are supported.

    """
    def __init__(self, backend: Backend) -> None:
        self.backend = backend
        self._output_backend = backend
        self._polygon_buffer = PolygonBuffer()
        self.page = Page(1189, 841)
        self.properties = Properties()
        self.is_pen_down = False
        self.is_absolute_mode = True
        self.is_polygon_mode = False
        self.has_merge_control = False
        self._user_location = NULLVEC2
        self._pen_state_stack: list[bool] = []

    @property
    def user_location(self) -> Vec2:
        """Returns the current pen location as point in the user coordinate system."""
        return self._user_location

    @property
    def page_location(self) -> Vec2:
        """Returns the current pen location as page point in plotter units."""
        location = self.user_location
        return self.page.page_point(location.x, location.y)

    def setup_page(self, size_x: int, size_y: int):
        self.page = Page(size_x, size_y)

    def set_scaling_points(self, p1: Vec2, p2: Vec2) -> None:
        self.page.set_scaling_points(p1, p2)

    def set_scaling_points_relative_1(self, xp1: float, yp1: float) -> None:
        self.page.set_scaling_points_relative_1(xp1, yp1)

    def set_scaling_points_relative_2(
        self, xp1: float, yp1: float, xp2: float, yp2: float
    ) -> None:
        self.page.set_scaling_points_relative_2(xp1, yp1, xp2, yp2)

    def reset_scaling(self) -> None:
        self.page.reset_scaling()

    def set_point_factor(self, origin: Vec2, scale_x: float, scale_y: float) -> None:
        self.page.set_ucs(origin, scale_x, scale_y)

    def set_isotropic_scaling(
        self,
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
        left=0.5,
        bottom=0.5,
    ) -> None:
        self.page.set_isotropic_scaling(x_min, x_max, y_min, y_max, left, bottom)

    def set_anisotropic_scaling(
        self, x_min: float, x_max: float, y_min: float, y_max: float
    ) -> None:
        self.page.set_anisotropic_scaling(x_min, x_max, y_min, y_max)

    def set_merge_control(self, status: bool) -> None:
        self.has_merge_control = status

    def pen_up(self) -> None:
        self.is_pen_down = False

    def pen_down(self) -> None:
        self.is_pen_down = True

    def push_pen_state(self) -> None:
        self._pen_state_stack.append(self.is_pen_down)

    def pop_pen_state(self) -> None:
        if len(self._pen_state_stack):
            self.is_pen_down = self._pen_state_stack.pop()

    def move_to(self, location: Vec2) -> None:
        if self.is_absolute_mode:
            self.move_to_abs(location)
        else:
            self.move_to_rel(location)

    def move_to_abs(self, user_location: Vec2) -> None:
        self._user_location = user_location

    def move_to_rel(self, user_location: Vec2) -> None:
        self._user_location += user_location

    def set_absolute_mode(self) -> None:
        self.is_absolute_mode = True

    def set_relative_mode(self) -> None:
        self.is_absolute_mode = False

    def set_current_pen(self, index: int) -> None:
        self.properties.set_current_pen(index)

    def set_max_pen_count(self, index: int) -> None:
        self.properties.set_max_pen_count(index)

    def set_pen_width(self, index: int, width: float) -> None:
        self.properties.set_pen_width(index, width)

    def set_pen_color(self, index: int, color: RGB) -> None:
        self.properties.set_pen_color(index, color)

    def set_fill_type(self, fill_type: int, spacing: float, angle: float) -> None:
        if fill_type in (3, 4):  # adjust spacing between hatching lines
            spacing = max(self.page.scale_length(spacing))
        self.properties.set_fill_type(fill_type, spacing, angle)

    def enter_polygon_mode(self, status: int) -> None:
        self.is_polygon_mode = True
        self.backend = self._polygon_buffer
        if status == 0:
            self._polygon_buffer.reset(self.page_location)
        elif status == 1:
            self._polygon_buffer.close_path()

    def exit_polygon_mode(self) -> None:
        self.is_polygon_mode = False
        self._polygon_buffer.close_path()
        self.backend = self._output_backend

    def fill_polygon(self, fill_method: int) -> None:
        self.properties.set_fill_method(fill_method)
        self.plot_filled_polygon_buffer(self._polygon_buffer.get_paths())

    def edge_polygon(self) -> None:
        self.plot_outline_polygon_buffer(self._polygon_buffer.get_paths())

    def plot_polyline(self, points: Sequence[Vec2]):
        if not points:
            return
        if self.is_absolute_mode:
            self.plot_abs_polyline(points)
        else:
            self.plot_rel_polyline(points)

    def plot_abs_polyline(self, points: Sequence[Vec2]):
        # input coordinates are user coordinates
        if not points:
            return
        current_page_location = self.page_location
        self.move_to_abs(points[-1])  # user coordinates!
        if self.is_pen_down:
            # convert to page coordinates:
            points = self.page.page_points(points)
            # insert current page location as starting point:
            points.insert(0, current_page_location)
            # draw polyline in absolute page coordinates:
            self.backend.draw_polyline(self.properties, points)

    def plot_rel_polyline(self, points: Sequence[Vec2]):
        # input coordinates are user coordinates
        if not points:
            return
        # convert to absolute user coordinates:
        self.plot_abs_polyline(
            tuple(rel_to_abs_points_dynamic(self.user_location, points))
        )

    def plot_abs_circle(self, radius: float, chord_angle: float):
        # radius in user units
        if self.is_pen_down:
            center = self.user_location
            vertices = [
                center + Vec2.from_deg_angle(a, radius)
                for a in arc_angles(0, 360.0, chord_angle)
            ]
            # draw circle in absolute page coordinates:
            self.backend.draw_polyline(self.properties, vertices)

    def plot_abs_arc(self, center: Vec2, sweep_angle: float, chord_angle: float):
        start_point = self.user_location
        radius_vec = start_point - center
        radius = radius_vec.magnitude
        start_angle = radius_vec.angle_deg
        end_angle = start_angle + sweep_angle
        end_point = center + Vec2.from_deg_angle(end_angle, radius)

        self.move_to_abs(end_point)
        if self.is_pen_down:
            vertices = [
                center + Vec2.from_deg_angle(a, radius)
                for a in arc_angles(start_angle, sweep_angle, chord_angle)
            ]
            self.backend.draw_polyline(self.properties, vertices)

    def plot_rel_arc(self, center_rel: Vec2, sweep_angle: float, chord_angle: float):
        self.plot_abs_arc(center_rel + self.user_location, sweep_angle, chord_angle)

    def plot_abs_arc_three_points(self, inter: Vec2, end: Vec2, chord_angle: float):
        # input coordinates are user coordinates
        start = self.user_location
        circle = ConstructionCircle.from_3p(start, inter, end)
        center = circle.center
        start_angle = (start - center).angle_deg
        end_angle = (end - center).angle_deg
        inter_angle = (inter - center).angle_deg
        sweep_angle = sweeping_angle(start_angle, inter_angle, end_angle)
        self.plot_abs_arc(center, sweep_angle, chord_angle)

    def plot_rel_arc_three_points(self, inter: Vec2, end: Vec2, chord_angle: float):
        # input coordinates are user coordinates
        current = self.user_location
        self.plot_abs_arc_three_points(current + inter, current + end, chord_angle)

    def plot_abs_cubic_bezier(self, ctrl1: Vec2, ctrl2: Vec2, end: Vec2):
        # input coordinates are user coordinates
        current_page_location = self.page_location
        self.move_to_abs(end)  # user coordinates!
        if self.is_pen_down:
            # convert to page coordinates:
            ctrl1, ctrl2, end = self.page.page_points((ctrl1, ctrl2, end))
            # draw cubic bezier curve in absolute page coordinates:
            p = Path(current_page_location)
            p.curve4_to(end, ctrl1, ctrl2)
            self.backend.draw_paths(self.properties, [p], filled=False)

    def plot_rel_cubic_bezier(self, ctrl1: Vec2, ctrl2: Vec2, end: Vec2):
        # input coordinates are user coordinates
        ctrl1, ctrl2, end = rel_to_abs_points_static(
            self.user_location, (ctrl1, ctrl2, end)
        )
        self.plot_abs_cubic_bezier(ctrl1, ctrl2, end)

    def plot_filled_polygon_buffer(self, paths: Sequence[Path]):
        # input coordinates are page coordinates!
        self.backend.draw_paths(self.properties, paths, filled=True)

    def plot_outline_polygon_buffer(self, paths: Sequence[Path]):
        # input coordinates are page coordinates!
        self.backend.draw_paths(self.properties, paths, filled=False)


def rel_to_abs_points_dynamic(current: Vec2, points: Sequence[Vec2]) -> Iterator[Vec2]:
    """Returns the absolute location of increment points, each point is an increment
    of the previous point starting at the current pen location.
    """
    for point in points:
        current += point
        yield current


def rel_to_abs_points_static(current: Vec2, points: Sequence[Vec2]) -> Iterator[Vec2]:
    """Returns the absolute location of increment points, all points are relative
    to the current pen location.
    """
    for point in points:
        yield current + point


def arc_angles(start: float, sweep_angle: float, chord_angle: float) -> Iterator[float]:
    # clamp to 0.5 .. 180
    chord_angle = min(180.0, max(0.5, chord_angle))
    count = abs(round(sweep_angle / chord_angle))
    delta = sweep_angle / count
    for index in range(count + 1):
        yield start + delta * index


def sweeping_angle(start: float, intermediate: float, end: float) -> float:
    """Returns the sweeping angle from start angle to end angle passing the
    intermediate angle.
    """
    start = start % 360.0
    intermediate = intermediate % 360.0
    end = end % 360.0
    angle = end - start
    i_to_s = start - intermediate
    i_to_e = end - intermediate
    if math.isclose(abs(i_to_e) + abs(i_to_s), abs(angle)):
        return angle
    else:  # return complementary angle with opposite orientation
        if angle < 0:
            return 360.0 + angle
        else:
            return angle - 360.0



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
import pytest
import math
import pickle

# Import from 'ezdxf.math._vector' to test Python implementation
from ezdxf.math._vector import Vec3
from ezdxf.acc import USE_C_EXT

vec3_classes = [Vec3]

if USE_C_EXT:
    from ezdxf.acc.vector import Vec3 as CVec3

    vec3_classes.append(CVec3)


@pytest.fixture(params=vec3_classes)
def vec3(request):
    return request.param


def test_default_constructor(vec3):
    assert vec3() == (0, 0, 0)


def test_init_one_param(vec3):
    assert vec3((2, 3)) == (2, 3, 0)
    assert vec3((2, 3, 4)) == (2, 3, 4)


def test_invalid_one_param_init(vec3):
    with pytest.raises(TypeError):
        vec3(0)
    with pytest.raises(TypeError):
        vec3((0,))


def test_invalid_param_count(vec3):
    with pytest.raises(TypeError):
        vec3(1, 2, 3, 4)
    with pytest.raises(TypeError):
        vec3((1, 2, 3, 4))


def test_init_two_params(vec3):
    assert vec3(1, 2) == (1, 2, 0)


def test_init_three_params(vec3):
    assert vec3(1, 2, 3) == (1, 2, 3)


def test_immutable_attributes(vec3):
    v = vec3()
    with pytest.raises(AttributeError):
        v.x = 1.0
    with pytest.raises(AttributeError):
        v.y = 1.0
    with pytest.raises(AttributeError):
        v.z = 1.0


def test_from_angle(vec3):
    angle = math.radians(50)
    length = 3.0
    assert vec3.from_angle(angle, length) == (
        math.cos(angle) * length,
        math.sin(angle) * length,
        0,
    )

    length, angle = 7, 45
    v = vec3.from_deg_angle(angle, length)
    x = math.cos(math.radians(angle)) * length
    y = math.sin(math.radians(angle)) * length
    assert v == (x, y)


def test_usage_as_tuple(vec3):
    v = vec3(1, 2, 3)
    assert tuple(v) == (1, 2, 3)
    assert isinstance(v.xyz, tuple)
    assert v.xyz == (1, 2, 3)


def test_get_item_positive_index(vec3):
    v = vec3(1, 2, 3)
    assert v[0] == 1
    assert v[1] == 2
    assert v[2] == 3
    with pytest.raises(IndexError):
        _ = v[3]


@pytest.mark.parametrize("index", [-1, -2, -3])
def test_get_item_negative_index(index, vec3):
    with pytest.raises(IndexError):
        _ = vec3()[index]  # negative indices not supported


def test_get_item_does_not_support_slicing(vec3):
    with pytest.raises(TypeError):
        _ = vec3()[:2]  # slicing is not supported


def test_vec2(vec3):
    v2 = vec3(1, 2, 3).vec2
    assert len(v2) == 2
    assert v2 == (1, 2)


def test_round(vec3):
    v = vec3(1.123, 2.123, 3.123)
    v2 = v.round(1)
    assert v2 == (1.1, 2.1, 3.1)


def test_iter(vec3):
    assert sum(vec3(1, 2, 3)) == 6


def test_deep_copy(vec3):
    import copy

    v = vec3(1, 2, 3)
    l1 = [v, v, v]

    l2 = copy.copy(l1)
    assert l2[0] is l1[0], "Vec3 is immutable"

    l3 = copy.deepcopy(l1)
    assert l3[0] is l1[0], "Vec3 is immutable"


def test_get_angle(vec3):
    v = vec3(3, 3)
    assert math.isclose(v.angle_deg, 45)
    assert math.isclose(v.angle, math.radians(45))


def test_spatial_angle(vec3):
    v = vec3(3, 3, 0)
    assert math.isclose(v.spatial_angle_deg, 45)
    assert math.isclose(v.spatial_angle, math.radians(45))


def test_compare_vectors(vec3):
    v1 = vec3(1, 2, 3)
    assert v1 == (1, 2, 3)
    assert (1, 2, 3) == v1

    v2 = vec3(2, 3, 4)
    assert v2 > v1
    assert v1 < v2


def test_xy(vec3):
    assert vec3(1, 2, 3).xy == vec3(1, 2)


def test_is_null(vec3):
    v = vec3()
    assert v.is_null

    v1 = vec3(23.56678, 56678.56778, 2.56677) * (1.0 / 14.5667)
    v2 = vec3(23.56678, 56678.56778, 2.56677) / 14.5667
    assert (v2 - v1).is_null

    assert vec3(0, 0, 0).is_null


def test_is_not_null_default_abs_tol(vec3):
    assert vec3(1e-11, 0, 0).is_null is False


def test_is_null_default_abs_tol(vec3):
    assert vec3(1e-12, 0, 0).is_null is True


def test_bool(vec3):
    v = vec3()
    assert bool(v) is False

    v1 = vec3(23.56678, 56678.56778, 2.56677) * (1.0 / 14.5667)
    v2 = vec3(23.56678, 56678.56778, 2.56677) / 14.5667
    result = v2 - v1
    assert bool(result) is False
    # current rel_tol=1e-9
    assert not vec3(1e-8, 0, 0).is_null


def test_magnitude(vec3):
    v = vec3(3, 4, 5)
    assert math.isclose(abs(v), 7.0710678118654755)
    assert math.isclose(v.magnitude, 7.0710678118654755)
    assert vec3().magnitude == 0


def test_magnitude_square(vec3):
    v = vec3(3, 4, 5)
    assert math.isclose(v.magnitude_square, 50)
    assert vec3().magnitude_square == 0


def test_normalize(vec3):
    v = vec3(2, 0, 0)
    assert v.normalize() == (1, 0, 0)


def test_normalize_to_length(vec3):
    v = vec3(2, 0, 0)
    assert v.normalize(4) == (4, 0, 0)


def test_normalize_error(vec3):
    with pytest.raises(ZeroDivisionError):
        vec3().normalize()


def test_orthogonal_ccw(vec3):
    v = vec3(3, 4)
    assert v.orthogonal() == (-4, 3)


def test_orthogonal_cw(vec3):
    v = vec3(3, 4)
    assert v.orthogonal(False) == (4, -3)


def test_negative(vec3):
    v = vec3(2, 3, 4)
    assert -v == (-2, -3, -4)
    assert (-vec3()).is_null is True


def test_add_vector(vec3):
    assert vec3(2, 3, 4) + (7, 7, 7) == (9, 10, 11)


def test_iadd_vector(vec3):
    v = vec3(2, 3, 4)
    v += (7, 7, 7)
    assert v == (9, 10, 11)


def test_radd_vector(vec3):
    assert (7, 7, 7) + vec3(2, 3, 4) == (9, 10, 11)


def test_add_scalar_type_error(vec3):
    with pytest.raises(TypeError):
        vec3(2, 3, 4) + 3


def test_radd_scalar_type_error(vec3):
    with pytest.raises(TypeError):
        3 + vec3(2, 3, 4)


def test_iadd_scalar_type_error(vec3):
    v = vec3(2, 3, 4)
    with pytest.raises(TypeError):
        v += 3


def test_sub_vector(vec3):
    assert vec3(2, 3, 4) - (7, 7, 7) == (-5, -4, -3)


def test_isub_vector(vec3):
    v = vec3(2, 3, 4)
    v -= (7, 7, 7)
    assert v == (-5, -4, -3)


def test_rsub_vector(vec3):
    assert (7, 7, 7) - vec3(2, 3, 4) == (5, 4, 3)


def test_sub_scalar_type_error(vec3):
    with pytest.raises(TypeError):
        vec3(2, 3, 4) - 3


def test_rsub_scalar_vector_type_error(vec3):
    with pytest.raises(TypeError):
        7 - vec3(2, 3, 4)


def test_isub_scalar_type_error(vec3):
    v = vec3(2, 3, 4)
    with pytest.raises(TypeError):
        v -= 3


def test_mul_scalar(vec3):
    assert vec3(2, 3, 4) * 2 == (4, 6, 8)


def test_mul_tuple_type_error(vec3):
    with pytest.raises(TypeError):
        vec3(1, 0) * (1, 9)


def test_imul_scalar(vec3):
    v = vec3(2, 3, 4)
    v *= 2
    assert v == (4, 6, 8)


def test_imul_tuple_type_error(vec3):
    v = vec3(2, 3, 4)
    with pytest.raises(TypeError):
        v *= (1, 0)


def test_rmul_scalar(vec3):
    assert 2 * vec3(2, 3, 4) == (4, 6, 8)


def test_rmul_tuple_type_error(vec3):
    with pytest.raises(TypeError):
        (1, 9) * vec3(1, 0)


def test_div_scalar(vec3):
    assert vec3(2, 3, 4) / 2 == (1, 1.5, 2)
    with pytest.raises(ZeroDivisionError):
        vec3(1, 0) / 0


def test_idiv_scalar(vec3):
    v = vec3(2, 3, 4)
    v /= 2
    assert v == (1, 1.5, 2)


def test_div_tuple_type_error(vec3):
    with pytest.raises(TypeError):
        vec3(1, 0) / (1, 0)


def test_rdiv_scalar_type_error(vec3):
    with pytest.raises(TypeError):
        1 / vec3(1, 0)


def test_rdiv_tuple_type_error(vec3):
    with pytest.raises(TypeError):
        (1, 0) / vec3(1, 0)


def test_inplace_operations_do_not_mutate_vec3_inplace():
    v = Vec3(2, 3)
    v_check = v
    v += Vec3(7, 7)
    assert v_check is not v, "__iadd__ should not operate inplace"

    v_check = v
    v -= Vec3(7, 7)
    assert v_check is not v, "__isub__ should not operate inplace"

    v_check = v
    v *= 1
    assert v_check is not v, "__imul__ should not operate inplace"

    v_check = v
    v /= 1
    assert v_check is not v, "__itruediv__ should not operate inplace"


def test_dot_product(vec3):
    v1 = vec3(2, 7, 1)
    v2 = vec3(3, 9, 8)
    assert math.isclose(v1.dot(v2), 77)


def test_angle_deg(vec3):
    assert math.isclose(vec3(0, 1).angle_deg, 90)
    assert math.isclose(vec3(0, -1).angle_deg, -90)
    assert math.isclose(vec3(1, 1).angle_deg, 45)
    assert math.isclose(vec3(-1, 1).angle_deg, 135)


def test_angle_between(vec3):
    v1 = vec3(0, 1)
    v2 = vec3(1, 1)
    angle = v1.angle_between(v2)
    assert math.isclose(angle, math.pi / 4)
    # reverse order, same result
    angle = v2.angle_between(v1)
    assert math.isclose(angle, math.pi / 4)
    angle = v1.angle_between(vec3(0, -1))
    assert math.isclose(angle, math.pi)


@pytest.mark.parametrize(
    "v1, v2",
    [
        [(1, 0), (0, 0)],
        [(0, 0), (1, 0)],
        [(0, 0), (0, 0)],
    ],
)
def test_angle_between_null_vector(vec3, v1, v2):
    with pytest.raises(ZeroDivisionError):
        vec3(v1).angle_between(vec3(v2))


def test_angle_about(vec3):
    extrusion = vec3(0, 0, 1)
    a = vec3(1, 0, 0)
    b = vec3(1, 1, 0)
    assert math.isclose(a.angle_between(b), math.pi / 4)
    assert math.isclose(extrusion.angle_about(a, b), math.pi / 4)

    extrusion = vec3(0, 0, -1)
    assert math.isclose(a.angle_between(b), math.pi / 4)
    assert math.isclose(extrusion.angle_about(a, b), -math.pi / 4 % math.tau)

    extrusion = vec3(0, 0, 1)
    a = vec3(1, 1, 0)
    b = vec3(1, 1, 0)
    assert math.isclose(a.angle_between(b), 0, abs_tol=1e-5)
    assert math.isclose(extrusion.angle_about(a, b), 0)

    extrusion = vec3(0, 1, 0)
    a = vec3(1, 1, 0)
    b = vec3(0, 1, -1)
    assert math.isclose(a.angle_between(b), math.pi / 3, abs_tol=1e-5)
    c = a.cross(b)
    assert math.isclose(a.angle_between(b), c.angle_about(a, b))
    assert math.isclose(extrusion.angle_about(a, b), math.pi / 2)


def test_cross_product(vec3):
    v1 = vec3(2, 7, 9)
    v2 = vec3(3, 9, 1)
    assert v1.cross(v2) == (-74, 25, -3)


def test_rot_z(vec3):
    assert vec3(2, 2, 7).rotate_deg(90).isclose((-2, 2, 7))


def test_lerp(vec3):
    v1 = vec3(1, 1, 1)
    v2 = vec3(4, 4, 4)
    assert v1.lerp(v2, 0.5) == (2.5, 2.5, 2.5)
    assert v1.lerp(v2, 0) == (1, 1, 1)
    assert v1.lerp(v2, 1) == (4, 4, 4)


def test_replace(vec3):
    v = vec3(1, 2, 3)
    assert v.replace(x=7.0) == (7, 2, 3)
    assert v.replace(y=7.0) == (1, 7, 3)
    assert v.replace(z=7.0) == (1, 2, 7)
    assert v.replace(x=7.0, z=7.0) == (7, 2, 7)


def test_project(vec3):
    v = vec3(10, 0, 0)
    assert v.project((5, 0, 0)) == (5, 0, 0)
    assert v.project((5, 5, 0)) == (5, 0, 0)
    assert v.project((5, 5, 5)) == (5, 0, 0)

    v = vec3(10, 10, 0)
    assert v.project((10, 0, 0)).isclose((5, 5, 0))


def test_vec3_sum(vec3):
    assert vec3.sum([]).is_null is True
    assert vec3.sum([vec3(1, 1, 1)]) == (1, 1, 1)
    assert vec3.sum([vec3(1, 1, 1), (2, 2, 2)]) == (3, 3, 3)


def test_picklable(vec3):
    for v in [vec3(), vec3((1, 2.5, 3)), vec3(1, 2.5, 3)]:
        pickled_v = pickle.loads(pickle.dumps(v))
        assert v == pickled_v
        assert type(v) is type(pickled_v)


def test_is_equal(vec3):
    v1 = 1.23456789
    assert vec3(v1, v1, v1) == vec3(v1, v1, v1)
    assert vec3(v1, v1, v1) == (v1, v1, v1)
    assert vec3(v1, v1, 0) == vec3(v1, v1)
    assert vec3(v1, v1, 0) == (v1, v1)


def test_is_not_equal(vec3):
    v1 = 1.23456789
    v2 = 1.234567891
    assert vec3(v1, v1, v1) != vec3(v2, v2, v2)
    assert vec3(v1, v1, v1) != (v2, v2, v2)
    assert vec3(v1, v1, 0) != vec3(v2, v2)
    assert vec3(v1, v1, 0) != (v2, v2)
    assert vec3(v1, v1, 1) != vec3(v1, v1)
    assert vec3(v1, v1, 1) != (v1, v1)


@pytest.mark.parametrize(
    "a,b,rel_tol",
    [
        # maximum relative tolerance to be close
        (1.000001, 1.0000019, 1e-6),
        (10.000001, 10.0000019, 1e-7),
        (100.000001, 100.0000019, 1e-8),
        (1000.000001, 1000.0000019, 1e-9),
        (10000.000001, 10000.0000019, 1e-10),
        (100000.000001, 100000.0000019, 1e-11),
        (1000000.000001, 1000000.0000019, 1e-12),
    ],
)
def test_is_close_relative_tolerance(vec3, a, b, rel_tol):
    va = vec3(a, a, a)
    vb = vec3(b, b, b)
    assert va.isclose(vb, rel_tol=rel_tol)


@pytest.mark.parametrize(
    "a,b,rel_tol",
    [
        (1.000001, 1.0000019, 1e-7),
        (10.000001, 10.0000019, 1e-8),
        (100.000001, 100.0000019, 1e-9),
        (1000.000001, 1000.0000019, 1e-10),
        (10000.000001, 10000.0000019, 1e-11),
        (100000.000001, 100000.0000019, 1e-12),
        (1000000.000001, 1000000.0000019, 1e-13),
    ],
)
def test_is_not_close_relative_tolerance(vec3, a, b, rel_tol):
    va = vec3(a, a, a)
    vb = vec3(b, b, b)
    assert not va.isclose(vb, rel_tol=rel_tol)


@pytest.mark.parametrize(
    "a,b",
    [
        # default relative tolerance is 1e-9
        (10.00000001, 10.000000019),  # 1e-8
        (100.0000001, 100.00000019),  # 1e-7
        (1000.000001, 1000.0000019),  # 1e-6
        (10000.00001, 10000.000019),  # 1e-5
        (100000.0001, 100000.00019),  # 1e-4
    ],
)
def test_is_close_for_fixed_relative_tolerance(vec3, a, b):
    va = vec3(a, a, a)
    vb = vec3(b, b, b)
    assert va.isclose(vb, rel_tol=1e-9)


VALUES = [
    (10.000001, 10.0000019),
    (100.000001, 100.0000019),
    (1000.000001, 1000.0000019),
    (10000.000001, 10000.0000019),
    (100000.000001, 100000.0000019),
    (1000000.000001, 1000000.0000019),
]


@pytest.mark.parametrize("a,b", VALUES)
def test_is_close_absolute_tolerance(vec3, a, b):
    va = vec3(a, a, a)
    vb = vec3(b, b, b)
    assert va.isclose(vb, rel_tol=0, abs_tol=1e-6)


@pytest.mark.parametrize("a,b", VALUES)
def test_is_not_close_absolute_tolerance(vec3, a, b):
    va = vec3(a, a, a)
    vb = vec3(b, b, b)
    assert not va.isclose(vb, rel_tol=0, abs_tol=1e-7)


def test_loosing_floating_point_precision_for_big_values():
    # This values can be represented by a double without loss of precision
    assert not math.isclose(
        1_000_000_000.000001, 1_000_000_000.0000019, rel_tol=0, abs_tol=1e-7
    )

    # multiply by 10 and loose precision in the fractional part,
    # now the values are close enough to be equal:
    assert math.isclose(
        10_000_000_000.000001, 10_000_000_000.0000019, rel_tol=0, abs_tol=1e-7
    )



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
import pytest
from pathlib import Path

import ezdxf
from ezdxf.addons import odafc

NO_ODAFC = "ODA File Converter not installed or not found."


def dummy(name):
    with open(name, "wt") as fp:
        fp.write("Hello!\n")
    return name


# noinspection PyPep8Naming
class Test_Convert_ODAFC_Not_Required:
    def test_source_does_not_exist(self):
        with pytest.raises(FileNotFoundError):
            odafc.convert("xxx.dxf")

    def test_source_has_unsupported_extension(self, tmp_path):
        src = dummy(tmp_path / "xxx.err")
        with pytest.raises(odafc.UnsupportedFileFormat):
            odafc.convert(src)

    def test_destination_already_exist(self, tmp_path):
        src = dummy(tmp_path / "xxx.dxf")
        dummy(tmp_path / "xxx.dwg")
        with pytest.raises(FileExistsError):
            odafc.convert(src)

    def test_destination_folder_does_not_exist(self, tmp_path):
        src = dummy(tmp_path / "xxx.dxf")
        with pytest.raises(FileNotFoundError):
            odafc.convert(src, Path("ODAFC_XXX_FOLDER") / "xxx.dwg")

    def test_destination_has_unsupported_extension(self, tmp_path):
        src = dummy(tmp_path / "xxx.dxf")
        dst = tmp_path / "xxx.err"
        with pytest.raises(odafc.UnsupportedFileFormat):
            odafc.convert(src, dst)

    def test_invalid_DXF_version(self, tmp_path):
        src = dummy(tmp_path / "xxx.dxf")
        with pytest.raises(odafc.UnsupportedVersion):
            odafc.convert(src, version="ABC")


def dxf_r12(name):
    doc = ezdxf.new(dxfversion="R12")
    doc.modelspace().add_circle((0, 0), radius=1)
    doc.saveas(name)
    return name


# noinspection PyPep8Naming
@pytest.mark.skipif(not odafc.is_installed(), reason=NO_ODAFC)
class Test_Convert_ODAFC_Required:
    def test_dxf_to_dwg_same_stem(self, tmp_path):
        r12 = dxf_r12(tmp_path / "r12.dxf")
        odafc.convert(r12)
        assert r12.with_suffix(".dwg").exists() is True

    def test_dxf_to_dwg_new_stem(self, tmp_path):
        r12 = dxf_r12(tmp_path / "r12.dxf")
        dest = tmp_path / "r2018.dwg"
        odafc.convert(r12, dest)
        assert dest.exists() is True

    def test_upgrade_dxf_to_r2013(self, tmp_path):
        r12 = dxf_r12(tmp_path / "r12.dxf")
        dest = tmp_path / "r2013.dxf"
        odafc.convert(r12, dest, version="R2013")
        doc = ezdxf.readfile(dest)
        assert doc.acad_release == "R2013"


# noinspection PyPep8Naming
@pytest.mark.skipif(not odafc.is_installed(), reason=NO_ODAFC)
class Test_Export_And_Load_DWG:
    @pytest.fixture
    def doc(self):
        dxf = ezdxf.new(dxfversion="R12")
        dxf.modelspace().add_circle((0, 0), radius=1)
        return dxf

    def test_export_dwg(self, tmp_path, doc):
        dwg_path = tmp_path / "r2013_1.dwg"
        assert dwg_path.exists() is False

        odafc.export_dwg(doc, str(dwg_path), "R2013")
        assert dwg_path.exists() is True

    def test_read_dwg(self, tmp_path, doc):
        dwg_path = tmp_path / "r2013_2.dwg"
        assert dwg_path.exists() is False
        odafc.export_dwg(doc, str(dwg_path), "R2013")

        loaded_doc = odafc.readfile(str(dwg_path))
        assert loaded_doc.acad_release == "R2013"


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing_extensions import TypeAlias
import abc

from ezdxf.npshapes import NumpyPath2d
from .font_measurements import FontMeasurements

GlyphPath: TypeAlias = NumpyPath2d


class Glyphs(abc.ABC):
    font_measurements: FontMeasurements  # of the raw font
    space_width: float  # word spacing of the raw font

    @abc.abstractmethod
    def get_scaling_factor(self, cap_height: float) -> float:
        ...

    @abc.abstractmethod
    def get_text_length(
        self, text: str, cap_height: float, width_factor: float = 1.0
    ) -> float:
        ...

    def get_text_path(
        self, text: str, cap_height: float, width_factor: float = 1.0
    ) -> GlyphPath:
        glyph_paths = self.get_text_glyph_paths(text, cap_height, width_factor)
        if len(glyph_paths) == 0:
            return GlyphPath(None)
        return NumpyPath2d.concatenate(glyph_paths)

    @abc.abstractmethod
    def get_text_glyph_paths(
        self, text: str, cap_height: float, width_factor: float = 1.0
    ) -> list[GlyphPath]:
        ...



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Union, Iterable, Sequence, Optional
from pathlib import Path
import enum
import platform
import shutil
import subprocess
from uuid import uuid4
import tempfile

import ezdxf
from ezdxf.math import Matrix44, UVec, Vec3, Vec2
from ezdxf.render import MeshBuilder, MeshTransformer
from ezdxf.addons import meshex


class Operation(enum.Enum):
    union = 0
    difference = 1
    intersection = 2


CMD = "openscad"
UNION = Operation.union
DIFFERENCE = Operation.difference
INTERSECTION = Operation.intersection


def get_openscad_path() -> str:
    if platform.system() in ("Linux", "Darwin"):
        return CMD
    else:
        return ezdxf.options.get("openscad-addon", "win_exec_path").strip('"')


def is_installed() -> bool:
    """Returns ``True`` if OpenSCAD is installed.

    Searches on Windows the path stored in the options as "win_exec_path" in section
    "[openscad-addon]" which is "C:\\Program Files\\OpenSCAD\\openscad.exe" by default.

    Searches the "openscad" command on Linux and macOS.

    """
    if platform.system() in ("Linux", "Darwin"):
        return shutil.which(CMD) is not None
    return Path(get_openscad_path()).exists()


def run(script: str, exec_path: Optional[str] = None) -> MeshTransformer:
    """Executes the given `script` by OpenSCAD and returns the result mesh as
    :class:`~ezdxf.render.MeshTransformer`.

    Args:
        script: the OpenSCAD script as string
        exec_path: path to the executable as string or ``None`` to use the
            default installation path

    """
    if exec_path is None:
        exec_path = get_openscad_path()

    workdir = Path(tempfile.gettempdir())
    uuid = str(uuid4())
    # The OFF format is more compact than the default STL format
    off_path = workdir / f"ezdxf_{uuid}.off"
    scad_path = workdir / f"ezdxf_{uuid}.scad"

    scad_path.write_text(script)
    subprocess.call(
        [
            exec_path,
            "--quiet",
            "-o",
            str(off_path),
            str(scad_path),
        ]
    )
    # Remove the OpenSCAD temp file:
    scad_path.unlink(missing_ok=True)

    new_mesh = MeshTransformer()
    # Import the OpenSCAD result from OFF file:
    if off_path.exists():
        new_mesh = meshex.off_loads(off_path.read_text())

    # Remove the OFF temp file:
    off_path.unlink(missing_ok=True)
    return new_mesh


def str_matrix44(m: Matrix44) -> str:
    # OpenSCAD uses column major order!
    import numpy as np

    def cleanup(values: Iterable) -> Iterable:
        for value in values:
            if isinstance(value, np.float64):
                yield float(value)
            else:
                yield value

    s = ", ".join([str(list(cleanup(c))) for c in m.columns()])
    return f"[{s}]"


def str_polygon(
    path: Iterable[UVec],
    holes: Optional[Sequence[Iterable[UVec]]] = None,
) -> str:
    """Returns a ``polygon()`` command as string. This is a 2D command, all
    z-axis values of the input vertices are ignored and all paths and holes
    are closed automatically.

    OpenSCAD docs: https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Using_the_2D_Subsystem#polygon

    Args:
        path: exterior path
        holes: a sequences of one or more holes as vertices

    """

    def add_vertices(vertices):
        index = len(points)
        indices = []
        vlist = Vec2.list(vertices)
        if not vlist[0].isclose(vlist[-1]):
            vlist.append(vlist[0])

        for v in vlist:
            indices.append(index)
            points.append(f"  [{v.x:g}, {v.y:g}],")
            index += 1
        return indices

    points: list[str] = []
    paths = [add_vertices(path)]
    if holes is not None:
        for hole in holes:
            paths.append(add_vertices(hole))
    lines = ["polygon(points = ["]
    lines.extend(points)
    lines.append("],")
    if holes is not None:
        lines.append("paths = [")
        for indices in paths:
            lines.append(f"  {str(indices)},")
        lines.append("],")
    lines.append("convexity = 10);")
    return "\n".join(lines)


class Script:
    def __init__(self) -> None:
        self.data: list[str] = []

    def add(self, data: str) -> None:
        """Add a string."""
        self.data.append(data)

    def add_polyhedron(self, mesh: MeshBuilder) -> None:
        """Add `mesh` as ``polyhedron()`` command.

        OpenSCAD docs: https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Primitive_Solids#polyhedron

        """
        self.add(meshex.scad_dumps(mesh))

    def add_polygon(
        self,
        path: Iterable[UVec],
        holes: Optional[Sequence[Iterable[UVec]]] = None,
    ) -> None:
        """Add a ``polygon()`` command. This is a 2D command, all
        z-axis values of the input vertices are ignored and all paths and holes
        are closed automatically.

        OpenSCAD docs: https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Using_the_2D_Subsystem#polygon

        Args:
            path: exterior path
            holes: a sequence of one or more holes as vertices, or ``None`` for no holes

        """
        self.add(str_polygon(path, holes))

    def add_multmatrix(self, m: Matrix44) -> None:
        """Add a transformation matrix of type :class:`~ezdxf.math.Matrix44` as
        ``multmatrix()`` operation.

        OpenSCAD docs: https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#multmatrix

        """
        self.add(f"multmatrix(m = {str_matrix44(m)})")  # no pending ";"

    def add_translate(self, v: UVec) -> None:
        """Add a ``translate()`` operation.

        OpenSCAD docs: https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#translate

        Args:
            v: translation vector

        """
        vec = Vec3(v)
        self.add(f"translate(v = [{vec.x:g}, {vec.y:g}, {vec.z:g}])")

    def add_rotate(self, ax: float, ay: float, az: float) -> None:
        """Add a ``rotation()`` operation.

        OpenSCAD docs: https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#rotate

        Args:
            ax: rotation about the x-axis in degrees
            ay: rotation about the y-axis in degrees
            az: rotation about the z-axis in degrees

        """
        self.add(f"rotate(a = [{ax:g}, {ay:g}, {az:g}])")

    def add_rotate_about_axis(self, a: float, v: UVec) -> None:
        """Add a ``rotation()`` operation about the given axis `v`.

        OpenSCAD docs: https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#rotate

        Args:
            a: rotation angle about axis `v` in degrees
            v: rotation axis as :class:`ezdxf.math.UVec` object

        """
        vec = Vec3(v)
        self.add(f"rotate(a = {a:g}, v = [{vec.x:g}, {vec.y:g}, {vec.z:g}])")

    def add_scale(self, sx: float, sy: float, sz: float) -> None:
        """Add a ``scale()`` operation.

        OpenSCAD docs: https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#scale

        Args:
            sx: scaling factor for the x-axis
            sy: scaling factor for the y-axis
            sz: scaling factor for the z-axis

        """
        self.add(f"scale(v = [{sx:g}, {sy:g}, {sz:g}])")

    def add_resize(
        self,
        nx: float,
        ny: float,
        nz: float,
        auto: Optional[Union[bool, tuple[bool, bool, bool]]] = None,
    ) -> None:
        """Add a ``resize()`` operation.

        OpenSCAD docs: https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#resize

        Args:
            nx: new size in x-axis
            ny: new size in y-axis
            nz: new size in z-axis
            auto: If the `auto` argument is set to ``True``, the operation
                auto-scales any 0-dimensions to match. Set the `auto` argument
                as a  3-tuple of bool values to auto-scale individual axis.

        """
        main = f"resize(newsize = [{nx:g}, {ny:g}, {nz:g}]"
        if auto is None:
            self.add(main + ")")
            return
        elif isinstance(auto, bool):
            flags = str(auto).lower()
        else:
            flags = ", ".join([str(a).lower() for a in auto])
            flags = f"[{flags}]"
        self.add(main + f", auto = {flags})")

    def add_mirror(self, v: UVec) -> None:
        """Add a ``mirror()`` operation.

        OpenSCAD docs: https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#mirror

        Args:
            v: the normal vector of a plane intersecting the origin through
                which to mirror the object

        """
        n = Vec3(v).normalize()
        self.add(f"mirror(v = [{n.x:g}, {n.y:g}, {n.z:g}])")

    def get_string(self) -> str:
        """Returns the OpenSCAD build script."""
        return "\n".join(self.data)


def boolean_operation(op: Operation, mesh1: MeshBuilder, mesh2: MeshBuilder) -> str:
    """Returns an `OpenSCAD`_ script to apply the given boolean operation to the
    given meshes.

    The supported operations are:

        - UNION
        - DIFFERENCE
        - INTERSECTION

    """
    assert isinstance(op, Operation), "enum of type Operation expected"
    script = Script()
    script.add(f"{op.name}() {{")
    script.add_polyhedron(mesh1)
    script.add_polyhedron(mesh2)
    script.add("}")
    return script.get_string()



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.tools.binarydata import BitStream, EndOfBufferError


def test_read_bit():
    data = b"\xaa"
    bs = BitStream(data)
    assert bs.read_bit() == 1
    assert bs.read_bit() == 0
    assert bs.read_bit() == 1
    assert bs.read_bit() == 0
    assert bs.read_bit() == 1
    assert bs.read_bit() == 0
    assert bs.read_bit() == 1
    assert bs.read_bit() == 0
    with pytest.raises(EndOfBufferError):
        _ = bs.read_bit()


def test_read_bits():
    data = b"\x0f\x0f"
    bs = BitStream(data)
    assert bs.read_bits(4) == 0
    assert bs.read_bits(2) == 3
    assert bs.read_bits(3) == 6
    assert bs.read_bits(4) == 1
    assert bs.read_bits(3) == 7


def test_read_unsigned_byte():
    data = b"\x0f\x0f"
    bs = BitStream(data)
    assert bs.read_bits(4) == 0
    assert bs.read_unsigned_byte() == 0xF0


def test_read_signed_byte():
    data = b"\x0f\xf0"
    bs = BitStream(data)
    assert bs.read_bits(4) == 0
    assert bs.read_signed_byte() == -1


def test_read_unsigned_short():
    # little endian!
    data = b"\x0c\xda\xb0"
    bs = BitStream(data)
    assert bs.read_bits(4) == 0
    assert bs.read_unsigned_short() == 0xABCD
    assert bs.read_bits(4) == 0


def test_read_aligned_unsigned_short():
    # little endian!
    data = b"\x00\xcd\xab"
    bs = BitStream(data)
    assert bs.read_unsigned_byte() == 0
    assert bs.read_unsigned_short() == 0xABCD


def test_read_unsigned_long():
    # little endian!
    data = b"\x00\x0e\xfc\xda\xb0"
    bs = BitStream(data)
    assert bs.read_bits(4) == 0
    assert bs.read_unsigned_long() == 0xABCDEF00
    assert bs.read_bits(4) == 0


def test_read_aligned_unsigned_long():
    # little endian!
    data = b"\x00\xef\xcd\xab"
    bs = BitStream(data)
    assert bs.read_unsigned_long() == 0xABCDEF00


def test_read_bitshort():
    bs = BitStream(b"\xe0")
    assert bs.read_bit_short() == 256  # 11
    assert bs.read_bit_short() == 0  # 10
    bs = BitStream(b"\x00\xff\xff")
    bs.read_bits(6)
    assert bs.read_bit_short() == -1
    assert BitStream(b"\x7f\x00").read_bit_short() == 252


def test_read_signed_modular_chars():
    bs = BitStream(
        bytes(
            [
                0b11101001,
                0b10010111,
                0b11100110,
                0b00110101,
                0b10000010,
                0b00100100,
                0b10000101,
                0b01001011,
            ]
        )
    )
    mc = bs.read_signed_modular_chars()
    assert mc == 112823273
    mc = bs.read_signed_modular_chars()
    assert mc == 4610
    mc = bs.read_signed_modular_chars()
    assert mc == -1413


def test_read_unsigned_modular_chars():
    bs = BitStream(
        bytes(
            [
                0b11101001,
                0b10010111,
                0b11100110,
                0b00110101,
                0b10000010,
                0b00100100,
                0b10000101,
                0b01001011,
            ]
        )
    )
    mc = bs.read_unsigned_modular_chars()
    assert mc == 112823273
    mc = bs.read_unsigned_modular_chars()
    assert mc == 4610
    mc = bs.read_unsigned_modular_chars()
    assert mc == 9605


def test_read_modular_shorts():
    bs = BitStream(
        bytes(
            [
                0b00110001,
                0b11110100,
                0b10001101,
                0b00000000,
            ]
        )
    )
    ms = bs.read_modular_shorts()
    assert ms == 4650033


def test_read_object_type():
    assert BitStream(bytes([0b00000000, 0b01000000])).read_object_type() == 1
    assert (
        BitStream(bytes([0b01000000, 0b01000000])).read_object_type()
        == 1 + 0x1F0
    )
    assert (
        BitStream(
            bytes([0b10000000, 0b01000000, 0b01000000])
        ).read_object_type()
        == 257
    )
    assert (
        BitStream(
            bytes([0b11000000, 0b01000000, 0b01000000])
        ).read_object_type()
        == 257
    )


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2018-2020 Manfred Moitzi
# License: MIT License
from typing import cast
import pytest
import math
from unittest.mock import MagicMock

import ezdxf

from ezdxf.audit import Auditor
from ezdxf.lldxf import const
from ezdxf.lldxf.tagwriter import TagCollector
from ezdxf.lldxf.tags import Tags
from ezdxf.entities.mline import MLineVertex, MLine, MLineStyle
from ezdxf.math import Matrix44, Vec3
from ezdxf.protocols import SupportsVirtualEntities, query_virtual_entities


# noinspection PyUnresolvedReferences
class TestMLine:
    @pytest.fixture(scope="class")
    def msp(self):
        return ezdxf.new().modelspace()

    @pytest.fixture
    def mline_mock_update_geometry(self):
        mline = MLine()
        mline.update_geometry = MagicMock()
        return mline

    def test_unbounded_mline(self):
        mline = MLine()
        assert mline.dxf.style_handle is None
        assert mline.dxf.style_name == "Standard"
        assert mline.style is None

    def test_generic_mline(self, msp):
        mline = msp.add_mline()
        assert mline.dxftype() == "MLINE"
        assert mline.dxf.style_name == "Standard"
        assert mline.dxf.count == 0
        assert mline.dxf.start_location == (0, 0, 0)

    def test_set_justification(self, mline_mock_update_geometry):
        mline = mline_mock_update_geometry
        mline.set_justification(mline.BOTTOM)
        assert mline.dxf.justification == mline.BOTTOM
        mline.update_geometry.assert_called_once()

    def test_set_scale_factor(self, mline_mock_update_geometry):
        mline = mline_mock_update_geometry
        mline.set_scale_factor(17)
        assert mline.dxf.scale_factor == 17
        mline.update_geometry.assert_called_once()

    def test_close_state(self, mline_mock_update_geometry):
        mline = mline_mock_update_geometry
        assert mline.is_closed is False
        mline.close(True)
        assert mline.is_closed is True
        mline.update_geometry.assert_called_once()

    def test_point_count_management(self):
        mline = MLine()
        mline.load_vertices(Tags.from_text(VTX_2))
        assert len(mline.vertices) == 2
        assert len(mline) == 2
        assert mline.dxf.count == 2, "should be a callback to __len__()"

    def test_add_first_vertex(self):
        mline = MLine()
        mline.extend([(0, 0, 0)])
        assert mline.start_location() == (0, 0, 0)
        assert len(mline) == 1

    def test_add_two_vertices(self, msp):
        # MLineStyle is required
        mline = msp.add_mline([(0, 0), (10, 0)])
        assert mline.start_location() == (0, 0, 0)
        assert len(mline) == 2
        assert mline.vertices[0].line_direction.isclose((1, 0))
        assert mline.vertices[0].miter_direction.isclose((0, 1))
        assert mline.vertices[1].line_direction.isclose(
            (1, 0)
        ), "continue last line segment"
        assert mline.vertices[1].miter_direction.isclose((0, 1))

    def test_x_rotation(self, msp):
        mline = msp.add_mline([(0, 5), (10, 5)])
        m = Matrix44.x_rotate(math.pi / 2)
        mline.transform(m)
        assert mline.start_location().isclose((0, 0, 5))
        assert mline.dxf.extrusion.isclose((0, -1, 0))
        assert mline.dxf.scale_factor == 1

    def test_translate(self, msp):
        mline = msp.add_mline([(0, 5), (10, 5)])
        m = Matrix44.translate(1, 1, 1)
        mline.transform(m)
        assert mline.start_location().isclose((1, 6, 1))
        assert mline.dxf.scale_factor == 1

    def test_uniform_scale(self, msp):
        mline = msp.add_mline([(0, 5), (10, 5)])
        m = Matrix44.scale(2, 2, 2)
        mline.transform(m)
        assert mline.start_location().isclose((0, 10, 0))
        assert mline.dxf.scale_factor == 2

    def test_non_uniform_scale(self, msp):
        mline = msp.add_mline([(1, 2, 3), (3, 4, 3)])
        m = Matrix44.scale(2, 1, 3)
        mline.transform(m)
        assert mline.start_location().isclose((2, 2, 9))
        assert mline.dxf.scale_factor == 1, "ignore non-uniform scaling"

    def test_support_virtual_entities_protocol(self, msp):
        mline = msp.add_mline([(1, 2, 3), (3, 4, 3)])
        assert isinstance(mline, SupportsVirtualEntities)
        assert len(query_virtual_entities(mline)) > 0

    def test_virtual_sub_entities_source_tracking(self, msp):
        mline = msp.add_mline([(1, 2, 3), (3, 4, 3)])
        result = set(e.source_of_copy for e in mline.virtual_entities())
        assert len(result) == 1, "only one source of copy expected"
        assert mline in result, "mline should be the source of copy"


class TestMLineStyle:
    @pytest.fixture(scope="class")
    def doc(self):
        return ezdxf.new()

    def test_standard_mline_style(self, doc):
        mline_style = cast("MLineStyle", doc.mline_styles.get("Standard"))
        assert mline_style.dxftype() == "MLINESTYLE"

        elements = mline_style.elements
        assert len(elements) == 2
        assert elements[0].offset == 0.5
        assert elements[0].color == 256
        assert elements[0].linetype == "BYLAYER"
        assert elements[1].offset == -0.5
        assert elements[1].color == 256
        assert elements[1].linetype == "BYLAYER"

    def test_set_defined_style(self, doc):
        style = doc.mline_styles.new("DefinedStyle")
        mline = doc.modelspace().add_mline()
        mline.set_style("DefinedStyle")
        assert mline.dxf.style_name == "DefinedStyle"
        assert mline.dxf.style_handle == style.dxf.handle

    def test_set_undefined_style(self, doc):
        mline = doc.modelspace().add_mline()
        with pytest.raises(const.DXFValueError):
            mline.set_style("UndefinedStyle")

    def test_ordered_indices(self):
        style = MLineStyle()
        style.elements.append(5)  # top order
        style.elements.append(-5)  # bottom border
        style.elements.append(0)
        style.elements.append(1)
        assert style.ordered_indices() == [1, 2, 3, 0]

    def test_invalid_element_count(self, doc):
        style = doc.mline_styles.new("InvalidMLineStyle")
        assert len(style.elements) == 0
        auditor = Auditor(doc)
        style.audit(auditor)
        assert auditor.has_errors is True, "invalid element count"


class TestMLineVertex:
    def test_load_tags(self):
        tags = Tags.from_text(VTX_1)
        vtx = MLineVertex.load(tags)
        assert isinstance(vtx.location, Vec3)
        assert vtx.location == (0, 0, 0)
        assert vtx.line_direction == (1, 0, 0)
        assert vtx.miter_direction == (0, 1, 0)
        assert len(vtx.line_params) == 3
        p1, p2, p3 = vtx.line_params
        assert p1 == (0.5, 0.0)
        assert p2 == (0.0, 0.0)
        assert p3 == (-0.5, 0.0)
        assert len(vtx.fill_params) == 3
        assert sum(len(p) for p in vtx.fill_params) == 0

    def test_new(self):
        vtx = MLineVertex.new(
            (1, 1),
            (1, 0),
            (0, 1),
            [(0.5, 0), (0, 0)],
            [tuple(), tuple()],
        )
        assert vtx.location == (1, 1, 0)
        assert vtx.line_direction == (1, 0, 0)
        assert vtx.miter_direction == (0, 1, 0)
        assert len(vtx.line_params) == 2
        p1, p2 = vtx.line_params
        assert p1 == (0.5, 0)
        assert p2 == (0, 0)
        assert len(vtx.fill_params) == 2
        assert sum(len(p) for p in vtx.fill_params) == 0

    def test_export_dxf(self):
        t = Tags.from_text(VTX_1)
        vtx = MLineVertex.load(t)
        collector = TagCollector()
        vtx.export_dxf(collector)

        tags = Tags(collector.tags)
        assert tags[0] == (11, vtx.location[0])
        assert tags[1] == (21, vtx.location[1])
        assert tags[2] == (31, vtx.location[2])

        assert tags[3] == (12, vtx.line_direction[0])
        assert tags[4] == (22, vtx.line_direction[1])
        assert tags[5] == (32, vtx.line_direction[2])

        assert tags[6] == (13, vtx.miter_direction[0])
        assert tags[7] == (23, vtx.miter_direction[1])
        assert tags[8] == (33, vtx.miter_direction[2])

        # line- and fill parameters
        assert tags[9:] == t[3:]


class TestMLineAudit:
    @pytest.fixture(scope="class")
    def doc(self):
        d = ezdxf.new()
        new_style = d.mline_styles.new("NewStyle1")
        new_style.elements.append(0.5)
        new_style.elements.append(0)
        return d

    @pytest.fixture(scope="class")
    def msp(self, doc):
        return doc.modelspace()

    @pytest.fixture
    def auditor(self, doc):
        return Auditor(doc)

    @pytest.fixture
    def mline1(self, msp):
        return msp.add_mline([(0, 0), (1, 1)])

    def test_valid_mline(self, mline1, auditor):
        mline1.audit(auditor)
        assert auditor.has_errors is False
        assert auditor.has_fixes is False

    def test_fix_invalid_style_name(self, mline1, auditor):
        mline1.dxf.style_name = "test"
        mline1.audit(auditor)
        assert mline1.dxf.style_name == "Standard"
        assert auditor.has_fixes is False, "silent fix"

    def test_fix_invalid_style_handle(self, mline1, auditor):
        mline1.dxf.style_name = "test"
        mline1.dxf.style_handle = "0"
        mline1.audit(auditor)
        assert mline1.dxf.style_name == "Standard"
        assert (
            mline1.dxf.style_handle
            == auditor.doc.mline_styles["Standard"].dxf.handle
        )
        assert auditor.has_fixes is True

    def test_fix_invalid_style_handle_by_name(self, mline1, doc, auditor):
        new_style = doc.mline_styles.get("NewStyle1")
        mline1.dxf.style_name = "NewStyle1"
        mline1.dxf.style_handle = "0"
        mline1.audit(auditor)
        assert mline1.dxf.style_name == new_style.dxf.name
        assert mline1.dxf.style_handle == new_style.dxf.handle
        assert auditor.has_fixes is True

    def test_fix_invalid_line_direction(self, mline1, auditor):
        mline1.vertices[0].line_direction = (0, 0, 0)
        mline1.audit(auditor)
        assert auditor.has_fixes is True

    def test_fix_invalid_miter_direction(self, mline1, auditor):
        mline1.vertices[0].miter_direction = (0, 0, 0)
        mline1.audit(auditor)
        assert auditor.has_fixes is True

    def test_fix_invalid_line_parameters(self, mline1, auditor):
        mline1.vertices[0].line_params = []
        mline1.audit(auditor)
        assert auditor.has_fixes is True


VTX_1 = """11
0.0
21
0.0
31
0.0
12
1.0
22
0.0
32
0.0
13
0.0
23
1.0
33
0.0
74
2
41
0.5
41
0.0
75
0
74
2
41
0.0
41
0.0
75
0
74
2
41
-0.5
41
0.0
75
0
"""

VTX_2 = """11
0.0
21
0.0
31
0.0
11
10.0
21
0.0
31
0.0
"""



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License

import pytest
import copy
import math
from ezdxf.math import Vec3, Matrix44
from ezdxf.lldxf.const import DXFValueError, DXFStructureError, DXFTypeError
from ezdxf.lldxf.extendedtags import ExtendedTags
from ezdxf.lldxf.types import dxftag, tuples_to_tags
from ezdxf.lldxf.tags import (
    Tags,
    find_begin_and_end_of_encoded_xdata_tags,
    NotFoundException,
)
from ezdxf.entities.xdata import (
    XData,
    XDataUserDict,
    XDataUserList,
    transform_xdata_tags,
)
from ezdxf.lldxf.repair import filter_invalid_xdata_group_codes


class TagWriter:
    """Mockup"""

    def __init__(self):
        self.tags = []

    def write_tags(self, tags):
        self.tags.append(tags)


@pytest.fixture
def xdata():
    xtags = ExtendedTags.from_text(
        """0
DXFENTITY
1001
MOZMAN
1000
DataStr1
1000
DataStr2
1040
3.14
1001
ACAD
1000
AutoDesk
1000
Revit"""
    )
    return XData(xtags.xdata)


def test_empty_xdata():
    xdata = XData()
    assert len(xdata) == 0
    tagwriter = TagWriter()
    xdata.export_dxf(tagwriter)
    assert len(tagwriter.tags) == 0


def test_add_data(xdata):
    assert "XXX" not in xdata
    xdata.add("XXX", [(1000, "Extended Data String")])
    assert "XXX" in xdata


def test_safe_init():
    tags = ExtendedTags.from_text(
        """0
DXFENTITY
1001
MOZMAN
1000
DataStr1
1000
DataStr2
1007
XXX
1040
3.14
1001
ACAD
1000
AutoDesk
1000
Revit"""
    )

    xdata = XData.safe_init(tags.xdata)
    data = xdata.get("MOZMAN")
    assert (
        any(tag.code == 1007 for tag in data) is False
    ), "invalid group code 1007 should be removed"
    assert "ACAD" in xdata


@pytest.mark.parametrize("group_code", [1, 0, 1007])
def test_prevent_adding_invalid_data(group_code):
    xdata = XData()
    with pytest.raises(DXFValueError):
        xdata.add("INVALID", [(group_code, "String")])
    assert "INVALID" not in xdata


def test_get_data(xdata):
    assert "MOZMAN" in xdata
    tags = xdata.get("MOZMAN")
    assert len(tags) == 4
    assert tags[0] == (1001, "MOZMAN")
    assert tags[1] == (1000, "DataStr1")
    assert tags[2] == (1000, "DataStr2")
    assert tags[3] == (1040, 3.14)

    with pytest.raises(DXFValueError):
        _ = xdata.get("XXX")


def test_replace_xdata(xdata):
    assert len(xdata) == 2
    xdata.add("MOZMAN", [(1000, "XXX")])
    assert len(xdata) == 2, "replace existing XDATA"

    tags = xdata.get("MOZMAN")
    assert tags[0] == (1001, "MOZMAN")
    assert tags[1] == (1000, "XXX")


def test_discard_xdata(xdata):
    assert len(xdata) == 2
    xdata.discard("MOZMAN")
    assert len(xdata) == 1
    assert "MOZMAN" not in xdata
    # ignore none existing appids
    xdata.discard("xxx")
    assert len(xdata) == 1


def test_cloning(xdata):
    xdata2 = copy.deepcopy(xdata)
    assert len(xdata2) == 2

    assert "MOZMAN" in xdata2
    tags = xdata2.get("MOZMAN")
    assert len(tags) == 4
    assert tags[0] == (1001, "MOZMAN")
    assert tags[1] == (1000, "DataStr1")
    assert tags[2] == (1000, "DataStr2")
    assert tags[3] == (1040, 3.14)
    xdata2.discard("MOZMAN")
    assert "MOZMAN" in xdata
    assert "MOZMAN" not in xdata2


def test_dxf_export(xdata):
    tagwriter = TagWriter()
    xdata.export_dxf(tagwriter)
    result = tagwriter.tags
    assert len(result) == 2
    # appids in original order
    assert result[0][0] == (1001, "MOZMAN")
    assert result[1][0] == (1001, "ACAD")


def test_has_xdata_list(xdata):
    assert xdata.has_xlist("ACAD", "DSTYLE") is False


def test_get_xlist_exception(xdata):
    with pytest.raises(DXFValueError):
        _ = xdata.get_xlist("ACAD", "DSTYLE")


def test_set_xdata_list(xdata):
    xdata.set_xlist("ACAD", "DSTYLE", [(1070, 1), (1000, "String")])
    data = xdata.get_xlist("ACAD", "DSTYLE")
    assert len(data) == 5
    assert data == [
        (1000, "DSTYLE"),
        (1002, "{"),
        (1070, 1),
        (1000, "String"),
        (1002, "}"),
    ]
    # add another list to ACAD
    xdata.set_xlist("ACAD", "MOZMAN", [(1070, 2), (1000, "mozman")])
    data = xdata.get_xlist("ACAD", "MOZMAN")
    assert len(data) == 5
    assert data == [
        (1000, "MOZMAN"),
        (1002, "{"),
        (1070, 2),
        (1000, "mozman"),
        (1002, "}"),
    ]
    data = xdata.get("ACAD")
    assert len(data) == 10 + 3


def test_discard_xdata_list(xdata):
    xdata.set_xlist("ACAD", "DSTYLE", [(1070, 1), (1000, "String")])
    xdata_list = xdata.get_xlist("ACAD", "DSTYLE")
    assert len(xdata_list) == 5
    xdata.discard_xlist("ACAD", "DSTYLE")
    with pytest.raises(DXFValueError):
        _ = xdata.get_xlist("ACAD", "DSTYLE")

    xdata.discard_xlist("ACAD", "DSTYLE")


def test_replace_xdata_list(xdata):
    xdata.set_xlist("ACAD", "DSTYLE", [(1070, 1), (1000, "String")])
    xdata_list = xdata.get_xlist("ACAD", "DSTYLE")
    assert len(xdata_list) == 5
    assert xdata_list == [
        (1000, "DSTYLE"),
        (1002, "{"),
        (1070, 1),
        (1000, "String"),
        (1002, "}"),
    ]
    xdata.set_xlist(
        "ACAD", "DSTYLE", [(1070, 2), (1000, "mozman"), (1000, "data")]
    )
    xdata_list = xdata.get_xlist("ACAD", "DSTYLE")
    assert len(xdata_list) == 6
    assert xdata_list == [
        (1000, "DSTYLE"),
        (1002, "{"),
        (1070, 2),
        (1000, "mozman"),
        (1000, "data"),
        (1002, "}"),
    ]
    # replace not existing list -> append list
    xdata.replace_xlist("ACAD", "MOZMAN", [(1070, 3), (1000, "new")])
    xdata_list = xdata.get_xlist("ACAD", "MOZMAN")
    assert len(xdata_list) == 5
    assert xdata_list == [
        (1000, "MOZMAN"),
        (1002, "{"),
        (1070, 3),
        (1000, "new"),
        (1002, "}"),
    ]
    data = xdata.get("ACAD")
    assert len(data) == 6 + 5 + 3


def test_poyline_with_xdata():
    xdata = XData(
        ExtendedTags(
            filter_invalid_xdata_group_codes(
                Tags.from_text(POLYLINE_WITH_INVALID_XDATA)
            )
        ).xdata
    )
    assert len(xdata) == 2
    assert len(xdata.get("AVE_ENTITY_MATERIAL")) == 27


def test_remove_all_none_xdata_group_codes():
    tags = Tags.from_text(POLYLINE_WITH_INVALID_XDATA)
    tags = list(filter_invalid_xdata_group_codes(tags))
    assert len(tags) == 32


POLYLINE_WITH_INVALID_XDATA = """  0
POLYLINE
  5
2A
  8
T-POLYFACE-3DS
 62
     3
 66
     1
 10
0.0
 20
0.0
 30
0.0
 70
    64
 71
     8
 72
    12
1001
AVE_FINISH
1002
{
1070
     0
1005
0
1002
}
1001
AVE_ENTITY_MATERIAL
1002
{
1000

1002
{
1071
        0
1070
     0
1070
     0
1002
{
1070
     0
1070
     0
1070
     0
1040
0.0
1002
}
1070
     0
1070
     0
1002
{
1002
}
1002
}
1002
{
1002
}
1002
{
1002
}
1011
0.0
1021
0.0
1031
0.0
1021
0.0
1031
0.0
1011
1.0
1021
0.0
1031
0.0
1021
0.0
1031
0.0
1011
0.0
1021
0.0
1031
0.0
1021
1.0
1031
0.0
1011
0.0
1021
0.0
1031
0.0
1021
0.0
1031
1.0
1002
}
"""

COLUMNS_SPEC = """1001
ACAD
1000
ACAD_MTEXT_COLUMN_INFO_BEGIN
1000
ACAD_MTEXT_COLUMN_INFO_END
"""

WITHOUT_END_TAG = """1000
ACAD_MTEXT_COLUMN_INFO_BEGIN
"""

WITHOUT_BEGIN_TAG = """1000
ACAD_MTEXT_COLUMN_INFO_END
"""


class TestEncodedXDATATags:
    def test_find_begin_and_end_of_column_info(self):
        start, end = find_begin_and_end_of_encoded_xdata_tags(
            "ACAD_MTEXT_COLUMN_INFO", Tags.from_text(COLUMNS_SPEC)
        )
        assert start, end == (1, 3)

    def test_raise_exception_if_not_found(self):
        with pytest.raises(NotFoundException):
            find_begin_and_end_of_encoded_xdata_tags(
                "MOZMAN", Tags.from_text(COLUMNS_SPEC)
            )

    def test_raise_exception_without_end_tag(self):
        with pytest.raises(DXFStructureError):
            find_begin_and_end_of_encoded_xdata_tags(
                "ACAD_MTEXT_COLUMN_INFO", Tags.from_text(WITHOUT_END_TAG)
            )

    def test_raise_exception_without_begin_tag(self):
        with pytest.raises(DXFStructureError):
            find_begin_and_end_of_encoded_xdata_tags(
                "ACAD_MTEXT_COLUMN_INFO", Tags.from_text(WITHOUT_BEGIN_TAG)
            )


LIST1 = """1001
EZDXF
1000
DefaultList
1002
{
1000
VALUE
1000
CONTENT
1002
}
"""

USER_LIST = """1001
MyAppID
1000
MyUserList
1002
{
1000
VALUE
1000
CONTENT
1002
}
"""


class TestXDataUserList:
    @pytest.fixture
    def list1(self):
        return Tags.from_text(LIST1)

    @pytest.fixture
    def user_list(self):
        return Tags.from_text(USER_LIST)

    def test_load_not_existing_list(self):
        xlist = XDataUserList(XData())
        assert len(xlist) == 0

    def test_load_existing_list(self, list1):
        xlist = XDataUserList(XData([list1]))
        assert len(xlist) == 2
        assert xlist[0] == "VALUE"
        assert xlist[1] == "CONTENT"

    def test_load_user_list(self, user_list):
        xlist = XDataUserList(
            XData([user_list]), name="MyUserList", appid="MyAppID"
        )
        assert len(xlist) == 2
        assert xlist[0] == "VALUE"
        assert xlist[1] == "CONTENT"

    def test_list_like_getitem_interface(self, list1):
        xlist = XDataUserList(XData([list1]))
        assert len(xlist) == 2
        assert xlist[-1] == "CONTENT"
        assert xlist[:] == ["VALUE", "CONTENT"]

    def test_list_like_setitem_interface(self, list1):
        xlist = XDataUserList(XData([list1]))
        xlist[0] = 15
        assert xlist[0] == 15

    def test_list_like_insert_interface(self, list1):
        xlist = XDataUserList(XData([list1]))
        xlist.insert(0, 17)
        assert xlist[0] == 17
        assert len(xlist) == 3

    def test_list_like_delitem_interface(self, list1):
        xlist = XDataUserList(XData([list1]))
        del xlist[0]
        assert len(xlist) == 1
        assert xlist[0] == "CONTENT"

    def test_commit_creates_valid_XDATA(self):
        xlist = XDataUserList()
        xlist.extend(["String", Vec3(1, 2, 3), 3.1415, 256])
        xlist.commit()
        tags = xlist.xdata.get("EZDXF")
        assert tags == [
            dxftag(1001, "EZDXF"),
            dxftag(1000, "DefaultList"),
            dxftag(1002, "{"),
            dxftag(1000, "String"),
            dxftag(1010, (1, 2, 3)),
            dxftag(1040, 3.1415),
            dxftag(1071, 256),
            dxftag(1002, "}"),
        ]

    def test_commit_replaces_existing_XDATA(self, list1):
        xlist = XDataUserList(XData([list1]))
        xlist.clear()
        xlist.extend(["String", Vec3(1, 2, 3), 3.1415, 256])
        xlist.commit()
        tags = xlist.xdata.get("EZDXF")
        assert tags == [
            dxftag(1001, "EZDXF"),
            dxftag(1000, "DefaultList"),
            dxftag(1002, "{"),
            dxftag(1000, "String"),
            dxftag(1010, (1, 2, 3)),
            dxftag(1040, 3.1415),
            dxftag(1071, 256),
            dxftag(1002, "}"),
        ]

    def test_modify_existing_XDATA(self, list1):
        xlist = XDataUserList(XData([list1]))
        xlist[0] = 3.1415
        xlist[1] = 256
        xlist.commit()
        tags = xlist.xdata.get("EZDXF")
        assert tags == [
            dxftag(1001, "EZDXF"),
            dxftag(1000, "DefaultList"),
            dxftag(1002, "{"),
            dxftag(1040, 3.1415),
            dxftag(1071, 256),
            dxftag(1002, "}"),
        ]

    def test_modify_existing_user_list(self, user_list):
        xlist = XDataUserList(
            XData([user_list]), name="MyUserList", appid="MyAppID"
        )
        xlist[0] = 3.1415
        xlist[1] = 256
        xlist.commit()
        tags = xlist.xdata.get("MyAppID")
        assert tags == [
            dxftag(1001, "MyAppID"),
            dxftag(1000, "MyUserList"),
            dxftag(1002, "{"),
            dxftag(1040, 3.1415),
            dxftag(1071, 256),
            dxftag(1002, "}"),
        ]

    def test_context_manager_interface(self):
        xdata = XData()
        with XDataUserList(xdata) as xlist:
            xlist.extend(["String", Vec3(1, 2, 3), 3.1415, 256])

        tags = xdata.get("EZDXF")
        assert tags == [
            dxftag(1001, "EZDXF"),
            dxftag(1000, "DefaultList"),
            dxftag(1002, "{"),
            dxftag(1000, "String"),
            dxftag(1010, (1, 2, 3)),
            dxftag(1040, 3.1415),
            dxftag(1071, 256),
            dxftag(1002, "}"),
        ]

    @pytest.mark.parametrize("char", ["\n", "\r"])
    def test_invalid_line_break_characters_raise_exception(self, char):
        xlist = XDataUserList()
        xlist.append(char)
        with pytest.raises(DXFValueError):
            xlist.commit()

    def test_too_long_string_raise_exception(self):
        # The XDATA limit for group code 1000 is 255 characters
        xlist = XDataUserList()
        xlist.append("0123456789" * 26)
        with pytest.raises(DXFValueError):
            xlist.commit()


DICT1 = """1001
EZDXF
1000
DefaultDict
1002
{
1000
Key0
1000
StringValue
1000
Float
1040
3.1415
1002
}
"""

USER_DICT = """1001
MyAppID
1000
MyUserDict
1002
{
1000
MyKey
1000
MyString
1000
Int
1071
65535
1002
}
"""


class TestXDataUserDict:
    @pytest.fixture
    def dict1(self):
        return Tags.from_text(DICT1)

    @pytest.fixture
    def user_dict(self):
        return Tags.from_text(USER_DICT)

    def test_load_not_existing_dict(self):
        xdict = XDataUserDict(XData())
        assert len(xdict) == 0

    def test_load_existing_dict(self, dict1):
        xdict = XDataUserDict(XData([dict1]))
        assert len(xdict) == 2
        assert xdict["Key0"] == "StringValue"
        assert xdict["Float"] == 3.1415

    def test_load_user_dict(self, user_dict):
        xdict = XDataUserDict(
            XData([user_dict]), name="MyUserDict", appid="MyAppID"
        )
        assert len(xdict) == 2
        assert xdict["MyKey"] == "MyString"
        assert xdict["Int"] == 65535

    def test_dict_like_interface(self, dict1):
        xdict = XDataUserDict(XData([dict1]))
        assert len(xdict) == 2
        assert "Key0" in xdict
        assert list(xdict.keys()) == ["Key0", "Float"]
        assert list(xdict.values()) == ["StringValue", 3.1415]

    def test_dict_like_setitem_interface(self):
        xdict = XDataUserDict()
        xdict["Test"] = 17
        assert len(xdict) == 1
        assert "Test" in xdict

    @pytest.mark.parametrize("data", [0, 3.14, (1, 1), Vec3()])
    def test_setitem_accepts_only_strings(self, data):
        xdict = XDataUserDict()
        with pytest.raises(DXFTypeError):
            xdict[data] = "abc"

    def test_dict_like_del_interface(self, dict1):
        xdict = XDataUserDict(XData([dict1]))
        del xdict["Key0"]
        assert len(xdict) == 1
        xdict.clear()
        assert len(xdict) == 0

    def test_discard_non_existing_key(self):
        xdict = XDataUserDict()
        xdict.discard("XXX")
        assert len(xdict) == 0

    def test_commit_creates_valid_XDATA(self):
        xdict = XDataUserDict()
        xdict["Key0"] = "StringValue"
        xdict["Float"] = 3.1415
        xdict.commit()
        tags = xdict.xdata.get("EZDXF")
        assert tags == [
            dxftag(1001, "EZDXF"),
            dxftag(1000, "DefaultDict"),
            dxftag(1002, "{"),
            dxftag(1000, "Key0"),
            dxftag(1000, "StringValue"),
            dxftag(1000, "Float"),
            dxftag(1040, 3.1415),
            dxftag(1002, "}"),
        ]

    def test_context_manager_interface(self):
        xdata = XData()
        with XDataUserDict(xdata) as xdict:
            xdict["Key1"] = 256
            xdict["Key2"] = "Value2"

        tags = xdata.get("EZDXF")
        assert tags == [
            dxftag(1001, "EZDXF"),
            dxftag(1000, "DefaultDict"),
            dxftag(1002, "{"),
            dxftag(1000, "Key1"),
            dxftag(1071, 256),
            dxftag(1000, "Key2"),
            dxftag(1000, "Value2"),
            dxftag(1002, "}"),
        ]

    def test_update_from_dict(self):
        xdict = XDataUserDict()
        xdict.update({"a": 1, "b": "str"})
        assert xdict["a"] == 1
        assert xdict["b"] == "str"

    def test_update_supports_only_string_keys(self):
        xdict = XDataUserDict()
        with pytest.raises(DXFTypeError):
            xdict.update({0: 1})


class TestXDataTransformation:
    @pytest.fixture
    def tags(self):
        return Tags(
            tuples_to_tags(
                [
                    (1001, "EZDXF"),
                    (1010, (10, 20, 30)),
                    (1011, (11, 21, 31)),
                    (1012, (12, 22, 32)),
                    (1013, (13, 23, 33)),
                    (1041, 2),
                    (1042, 2),
                ]
            )
        )

    def test_move(self, tags):
        data = Tags(
            transform_xdata_tags(tags, Matrix44.translate(100, 100, 100))
        )
        # 1010 - fixed 3D point -> no transformation
        assert Vec3((10, 20, 30)).isclose(data[1].value)
        # 1011 - move, scale, rotate and mirror
        assert Vec3((111, 121, 131)).isclose(data[2].value)
        # 1012 - scale, rotate and mirror
        assert Vec3((12, 22, 32)).isclose(data[3].value)
        # 1013 - rotate and mirror
        assert Vec3((13, 23, 33)).isclose(data[4].value)
        # 1041 - scale distance
        assert math.isclose(2, data[5].value)
        # 1042 - scale factor
        assert math.isclose(2, data[6].value)

    def test_scale(self, tags):
        data = Tags(transform_xdata_tags(tags, Matrix44.scale(2, 2, 2)))
        # 1010 - fixed 3D point -> no transformation
        assert Vec3((10, 20, 30)).isclose(data[1].value)
        # 1011 - move, scale, rotate and mirror
        assert Vec3((22, 42, 62)).isclose(data[2].value)
        # 1012 - scale, rotate and mirror
        assert Vec3((24, 44, 64)).isclose(data[3].value)
        # 1013 - rotate and mirror
        assert Vec3((13, 23, 33)).isclose(data[4].value)
        # 1041 - scale distance
        assert math.isclose(4, data[5].value)
        # 1042 - scale factor
        assert math.isclose(4, data[6].value)

    def test_mirror_x(self, tags):
        data = Tags(transform_xdata_tags(tags, Matrix44.scale(-1, 1, 1)))
        # 1010 - fixed 3D point -> no transformation
        assert Vec3((10, 20, 30)).isclose(data[1].value)
        # 1011 - move, scale, rotate and mirror
        assert Vec3((-11, 21, 31)).isclose(data[2].value)
        # 1012 - scale, rotate and mirror
        assert Vec3((-12, 22, 32)).isclose(data[3].value)
        # 1013 - rotate and mirror
        assert Vec3((-13, 23, 33)).isclose(data[4].value)
        # 1041 - scale distance - BricsCAD transforms to -2 ???
        assert math.isclose(2, data[5].value)
        # 1042 - scale factor - BricsCAD transforms to -2 ???
        assert math.isclose(2, data[6].value)

    def test_scale_and_mirror_y(self, tags):
        data = Tags(transform_xdata_tags(tags, Matrix44.scale(2, -2, 2)))
        # 1011 - move, scale, rotate and mirror
        assert Vec3((22, -42, 62)).isclose(data[2].value)
        # 1012 - scale, rotate and mirror
        assert Vec3((24, -44, 64)).isclose(data[3].value)
        # 1013 - rotate and mirror
        assert Vec3((13, -23, 33)).isclose(data[4].value)
        # 1041 - scale distance
        assert math.isclose(4, data[5].value)
        # 1042 - scale factor
        assert math.isclose(4, data[6].value)

    def test_rotate(self, tags):
        data = Tags(transform_xdata_tags(tags, Matrix44.z_rotate(math.pi / 2)))
        # 1011 - move, scale, rotate and mirror
        assert Vec3((-21, 11, 31)).isclose(data[2].value)
        # 1012 - scale, rotate and mirror
        assert Vec3((-22, 12, 32)).isclose(data[3].value)
        # 1013 - rotate and mirror
        assert Vec3((-23, 13, 33)).isclose(data[4].value)
        # 1041 - scale distance
        assert math.isclose(2, data[5].value)
        # 1042 - scale factor
        assert math.isclose(2, data[6].value)


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf
from ezdxf.tools.text import MTextParser
from ezdxf.addons import MTextExplode

TEXT = r"{\fDIN OT|b0|i0|c0|p34;Abbruch Infoschild\PNotaus Einbindung\P\pxi-3,l4,t4;ehem.^IStaubsauger}"


def test_parse_text():
    result = list(MTextParser(TEXT))
    assert len(result) == 11


def test_explode_mtext():
    doc = ezdxf.new()
    msp = doc.modelspace()

    mtext = msp.add_mtext(TEXT)
    mtext.text = TEXT
    explode = MTextExplode(msp)
    explode.explode(mtext, destroy=True)
    assert len(msp) == 7


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations

import pathlib
from typing import Iterable, NamedTuple, Optional, Sequence
import os
import platform
import json
import logging

from pathlib import Path
from fontTools.ttLib import TTFont, TTLibError
from .font_face import FontFace
from . import shapefile, lff

logger = logging.getLogger("ezdxf")
WINDOWS = "Windows"
LINUX = "Linux"
MACOS = "Darwin"


WIN_SYSTEM_ROOT = os.environ.get("SystemRoot", "C:/Windows")
WIN_FONT_DIRS = [
    # AutoCAD and BricsCAD do not support fonts installed in the user directory:
    "~/AppData/Local/Microsoft/Windows/Fonts",
    f"{WIN_SYSTEM_ROOT}/Fonts",
]
LINUX_FONT_DIRS = [
    "/usr/share/fonts",
    "/usr/local/share/fonts",
    "~/.fonts",
    "~/.local/share/fonts",
    "~/.local/share/texmf/fonts",
]
MACOS_FONT_DIRS = ["/Library/Fonts/", "/System/Library/Fonts/"]
FONT_DIRECTORIES = {
    WINDOWS: WIN_FONT_DIRS,
    LINUX: LINUX_FONT_DIRS,
    MACOS: MACOS_FONT_DIRS,
}

DEFAULT_FONTS = [
    "ArialUni.ttf",  # for Windows
    "Arial Unicode.ttf",  # for macOS
    "Arial.ttf",  # for the case  "Arial Unicode" does not exist
    "DejaVuSansCondensed.ttf",  # widths of glyphs is similar to Arial
    "DejaVuSans.ttf",
    "LiberationSans-Regular.ttf",
    "OpenSans-Regular.ttf",
]
CURRENT_CACHE_VERSION = 2


class CacheEntry(NamedTuple):
    file_path: Path  # full file path e.g. "C:\Windows\Fonts\DejaVuSans.ttf"
    font_face: FontFace


GENERIC_FONT_FAMILY = {
    "serif": "DejaVu Serif",
    "sans-serif": "DejaVu Sans",
    "monospace": "DejaVu Sans Mono",
}


class FontCache:
    def __init__(self) -> None:
        # cache key is the lowercase ttf font name without parent directories
        # e.g. "arial.ttf" for "C:\Windows\Fonts\Arial.ttf"
        self._cache: dict[str, CacheEntry] = dict()

    def __contains__(self, font_name: str) -> bool:
        return self.key(font_name) in self._cache

    def __getitem__(self, item: str) -> CacheEntry:
        return self._cache[self.key(item)]

    def __setitem__(self, item: str, entry: CacheEntry) -> None:
        self._cache[self.key(item)] = entry

    def __len__(self):
        return len(self._cache)

    def clear(self) -> None:
        self._cache.clear()

    @staticmethod
    def key(font_name: str) -> str:
        return str(font_name).lower()

    def add_entry(self, font_path: Path, font_face: FontFace) -> None:
        self._cache[self.key(font_path.name)] = CacheEntry(font_path, font_face)

    def get(self, font_name: str, fallback: str) -> CacheEntry:
        try:
            return self._cache[self.key(font_name)]
        except KeyError:
            entry = self._cache.get(self.key(fallback))
            if entry is not None:
                return entry
            else:  # no fallback font available
                raise FontNotFoundError("no fonts available, not even fallback fonts")

    def find_best_match(self, font_face: FontFace) -> Optional[FontFace]:
        entry = self._cache.get(self.key(font_face.filename), None)
        if entry:
            return entry.font_face
        return self.find_best_match_ex(
            family=font_face.family,
            style=font_face.style,
            weight=font_face.weight,
            width=font_face.width,
            italic=font_face.is_italic,
        )

    def find_best_match_ex(
        self,
        family: str = "sans-serif",
        style: str = "Regular",
        weight: int = 400,
        width: int = 5,
        italic: Optional[bool] = False,
    ) -> Optional[FontFace]:
        # italic == None ... ignore italic flag
        family = GENERIC_FONT_FAMILY.get(family, family)
        entries = filter_family(family, self._cache.values())
        if len(entries) == 0:
            return None
        elif len(entries) == 1:
            return entries[0].font_face
        entries_ = filter_style(style, entries)
        if len(entries_) == 1:
            return entries_[0].font_face
        elif len(entries_):
            entries = entries_
        # best match by weight, italic, width
        # Note: the width property is used to prioritize shapefile types:
        # 1st .shx; 2nd: .shp; 3rd: .lff
        result = sorted(
            entries,
            key=lambda e: (
                abs(e.font_face.weight - weight),
                e.font_face.is_italic is not italic,
                abs(e.font_face.width - width),
            ),
        )
        return result[0].font_face

    def loads(self, s: str) -> None:
        cache: dict[str, CacheEntry] = dict()
        try:
            content = json.loads(s)
        except json.JSONDecodeError:
            raise IOError("invalid JSON file format")
        try:
            version = content["version"]
            content = content["font-faces"]
        except KeyError:
            raise IOError("invalid cache file format")
        if version == CURRENT_CACHE_VERSION:
            for entry in content:
                try:
                    file_path, family, style, weight, width = entry
                except ValueError:
                    raise IOError("invalid cache file format")
                path = Path(file_path)  # full path, e.g. "C:\Windows\Fonts\Arial.ttf"
                font_face = FontFace(
                    filename=path.name,  # file name without parent dirs, e.g. "Arial.ttf"
                    family=family,  # Arial
                    style=style,  # Regular
                    weight=weight,  # 400 (Normal)
                    width=width,  # 5 (Normal)
                )
                cache[self.key(path.name)] = CacheEntry(path, font_face)
        else:
            raise IOError("invalid cache file version")
        self._cache = cache

    def dumps(self) -> str:
        faces = [
            (
                str(entry.file_path),
                entry.font_face.family,
                entry.font_face.style,
                entry.font_face.weight,
                entry.font_face.width,
            )
            for entry in self._cache.values()
        ]
        data = {"version": CURRENT_CACHE_VERSION, "font-faces": faces}
        return json.dumps(data, indent=2)

    def print_available_fonts(self, verbose=False) -> None:
        for entry in self._cache.values():
            print(f"{entry.file_path}")
            if not verbose:
                continue
            font_type = entry.file_path.suffix.lower()
            ff = entry.font_face
            if font_type in (".shx", ".shp"):
                print(f"  Shape font file: '{ff.filename}'")
            elif font_type == ".lff":
                print(f"  LibreCAD font file: '{ff.filename}'")
            else:
                print(f"  TrueType/OpenType font file: '{ff.filename}'")
                print(f"  family: {ff.family}")
                print(f"  style: {ff.style}")
                print(f"  weight: {ff.weight}, {ff.weight_str}")
                print(f"  width: {ff.width}, {ff.width_str}")
        print(f"\nfound {len(self._cache)} fonts")


def filter_family(family: str, entries: Iterable[CacheEntry]) -> list[CacheEntry]:
    key = str(family).lower()
    return [e for e in entries if e.font_face.family.lower().startswith(key)]


def filter_style(style: str, entries: Iterable[CacheEntry]) -> list[CacheEntry]:
    key = str(style).lower()
    return [e for e in entries if key in e.font_face.style.lower()]


# TrueType and OpenType fonts:
# Note: CAD applications like AutoCAD/BricsCAD do not support OpenType fonts!
SUPPORTED_TTF_TYPES = {".ttf", ".ttc", ".otf"}
# Basic stroke-fonts included in CAD applications:
SUPPORTED_SHAPE_FILES = {".shx", ".shp", ".lff"}
NO_FONT_FACE = FontFace()
FALLBACK_SHAPE_FILES = ["txt.shx", "txt.shp", "iso.shx", "iso.shp"]
FALLBACK_LFF = ["standard.lff", "iso.lff", "simplex.lff"]


class FontNotFoundError(Exception):
    pass


class UnsupportedFont(Exception):
    pass


class FontManager:
    def __init__(self) -> None:
        self.platform = platform.system()
        self._font_cache: FontCache = FontCache()
        self._match_cache: dict[int, Optional[FontFace]] = dict()
        self._loaded_ttf_fonts: dict[str, TTFont] = dict()
        self._loaded_shape_file_glyph_caches: dict[str, shapefile.GlyphCache] = dict()
        self._loaded_lff_glyph_caches: dict[str, lff.GlyphCache] = dict()
        self._fallback_font_name = ""
        self._fallback_shape_file = ""
        self._fallback_lff = ""

    def print_available_fonts(self, verbose=False) -> None:
        self._font_cache.print_available_fonts(verbose=verbose)

    def has_font(self, font_name: str) -> bool:
        return font_name in self._font_cache

    def clear(self) -> None:
        self._font_cache = FontCache()
        self._loaded_ttf_fonts.clear()
        self._fallback_font_name = ""

    def fallback_font_name(self) -> str:
        fallback_name = self._fallback_font_name
        if fallback_name:
            return fallback_name
        fallback_name = DEFAULT_FONTS[0]
        for name in DEFAULT_FONTS:
            try:
                cache_entry = self._font_cache.get(name, fallback_name)
                fallback_name = cache_entry.file_path.name
                break
            except FontNotFoundError:
                pass
        self._fallback_font_name = fallback_name
        return fallback_name

    def fallback_shapefile(self) -> str:
        fallback_shape_file = self._fallback_shape_file
        if fallback_shape_file:
            return fallback_shape_file

        for name in FALLBACK_SHAPE_FILES:
            if name in self._font_cache:
                self._fallback_shape_file = name
                return name
        return ""

    def fallback_lff(self) -> str:
        fallback_lff = self._fallback_lff
        if fallback_lff:
            return fallback_lff

        for name in FALLBACK_SHAPE_FILES:
            if name in self._font_cache:
                self._fallback_shape_file = name
                return name
        return ""

    def get_ttf_font(self, font_name: str, font_number: int = 0) -> TTFont:
        try:
            return self._loaded_ttf_fonts[font_name]
        except KeyError:
            pass
        fallback_name = self.fallback_font_name()
        try:
            font = TTFont(
                self._font_cache.get(font_name, fallback_name).file_path,
                fontNumber=font_number,
            )
        except IOError as e:
            raise FontNotFoundError(str(e))
        except TTLibError as e:
            raise FontNotFoundError(str(e))
        self._loaded_ttf_fonts[font_name] = font
        return font

    def ttf_font_from_font_face(self, font_face: FontFace) -> TTFont:
        return self.get_ttf_font(Path(font_face.filename).name)

    def get_shapefile_glyph_cache(self, font_name: str) -> shapefile.GlyphCache:
        try:
            return self._loaded_shape_file_glyph_caches[font_name]
        except KeyError:
            pass
        fallback_name = self.fallback_shapefile()
        try:
            file_path = self._font_cache.get(font_name, fallback_name).file_path
        except KeyError:
            raise FontNotFoundError(f"shape font '{font_name}' not found")
        try:
            file = shapefile.readfile(str(file_path))
        except IOError:
            raise FontNotFoundError(f"shape file '{file_path}' not found")
        except shapefile.UnsupportedShapeFile as e:
            raise UnsupportedFont(f"unsupported font '{file_path}': {str(e)}")
        try:
            glyph_cache = shapefile.GlyphCache(file)
        except Exception:
            raise UnsupportedFont(f"can't create glyph-cache for font '{file_path}'.")
        self._loaded_shape_file_glyph_caches[font_name] = glyph_cache
        return glyph_cache

    def get_lff_glyph_cache(self, font_name: str) -> lff.GlyphCache:
        try:
            return self._loaded_lff_glyph_caches[font_name]
        except KeyError:
            pass
        fallback_name = self.fallback_lff()
        try:
            file_path = self._font_cache.get(font_name, fallback_name).file_path
        except KeyError:
            raise FontNotFoundError(f"LibreCAD font '{font_name}' not found")
        try:
            s = pathlib.Path(file_path).read_text(encoding="utf8")
            font = lff.loads(s)
        except IOError:
            raise FontNotFoundError(f"LibreCAD font file '{file_path}' not found")
        try:
            glyph_cache = lff.GlyphCache(font)
        except Exception:
            raise UnsupportedFont(f"can't create glyph-cache for font '{file_path}'.")

        self._loaded_lff_glyph_caches[font_name] = glyph_cache
        return glyph_cache

    def get_font_face(self, font_name: str) -> FontFace:
        cache_entry = self._font_cache.get(font_name, self.fallback_font_name())
        return cache_entry.font_face

    def find_best_match(
        self,
        family: str = "sans-serif",
        style: str = "Regular",
        weight=400,
        width=5,
        italic: Optional[bool] = False,
    ) -> Optional[FontFace]:
        key = hash((family, style, weight, width, italic))
        try:
            return self._match_cache[key]
        except KeyError:
            pass
        font_face = self._font_cache.find_best_match_ex(
            family, style, weight, width, italic
        )
        self._match_cache[key] = font_face
        return font_face

    def find_font_name(self, font_face: FontFace) -> str:
        """Returns the font file name of the font without parent directories
        e.g. "LiberationSans-Regular.ttf".
        """
        font_face = self._font_cache.find_best_match(font_face)  # type: ignore
        if font_face is None:
            font_face = self.get_font_face(self.fallback_font_name())
            return font_face.filename
        else:
            return font_face.filename

    def build(self, folders: Optional[Sequence[str]] = None, support_dirs=True) -> None:
        """Adds all supported font types located in the given `folders` to the font
        manager. If no directories are specified, the known font folders for Windows,
        Linux and macOS are searched by default, except `support_dirs` is ``False``.
        Searches recursively all subdirectories.

        The folders stored in the config SUPPORT_DIRS option are scanned recursively for
        .shx, .shp and .lff fonts, the basic stroke fonts included in CAD applications.

        """
        from ezdxf._options import options

        if folders:
            dirs = list(folders)
        else:
            dirs = FONT_DIRECTORIES.get(self.platform, LINUX_FONT_DIRS)
        if support_dirs:
            dirs = dirs + list(options.support_dirs)
        self.scan_all(dirs)

    def add_synonyms(self, synonyms: dict[str, str], reverse=True) -> None:
        font_cache = self._font_cache
        for font_name, synonym in synonyms.items():
            if not font_name in font_cache:
                continue
            if synonym in font_cache:
                continue
            cache_entry = font_cache[font_name]
            font_cache[synonym] = cache_entry
        if reverse:
            self.add_synonyms({v: k for k, v in synonyms.items()}, reverse=False)

    def scan_all(self, folders: Iterable[str]) -> None:
        for folder in folders:
            folder = folder.strip("'\"")  # strip quotes
            if not folder:
                continue
            try:
                self.scan_folder(Path(folder).expanduser())
            except PermissionError as e:
                print(str(e))
                continue

    def scan_folder(self, folder: Path):
        if not folder.exists():
            return
        for file in folder.iterdir():
            if file.is_dir():
                self.scan_folder(file)
                continue
            ext = file.suffix.lower()
            if ext in SUPPORTED_TTF_TYPES:
                try:
                    font_face = get_ttf_font_face(file)
                except Exception as e:
                    logger.warning(f"cannot open font '{file}': {str(e)}")
                else:
                    self._font_cache.add_entry(file, font_face)
            elif ext in SUPPORTED_SHAPE_FILES:
                font_face = get_shape_file_font_face(file)
                self._font_cache.add_entry(file, font_face)

    def dumps(self) -> str:
        return self._font_cache.dumps()

    def loads(self, s: str) -> None:
        self._font_cache.loads(s)


def normalize_style(style: str) -> str:
    if style in {"Book"}:
        style = "Regular"
    return style


def get_ttf_font_face(font_path: Path) -> FontFace:
    """The caller should catch ALL exception (see scan_folder function above) - strange
    things can happen when reading TTF files.
    """
    ttf = TTFont(font_path, fontNumber=0)
    names = ttf["name"].names
    family = ""
    style = ""
    for record in names:
        if record.nameID == 1:
            family = record.string.decode(record.getEncoding())
        elif record.nameID == 2:
            style = record.string.decode(record.getEncoding())
        if family and style:
            break

    try:
        os2_table = ttf["OS/2"]
    except Exception:  # e.g. ComickBook_Simple.ttf has an invalid "OS/2" table
        logger.info(f"cannot load OS/2 table of font '{font_path.name}'")
        weight = 400
        width = 5
    else:
        weight = os2_table.usWeightClass
        width = os2_table.usWidthClass
    return FontFace(
        filename=font_path.name,
        family=family,
        style=normalize_style(style),
        width=width,
        weight=weight,
    )


def get_shape_file_font_face(font_path: Path) -> FontFace:
    ext = font_path.suffix.lower()
    # Note: the width property is not defined in shapefiles and is used to
    # prioritize the shapefile types for find_best_match():
    # 1st .shx; 2nd: .shp; 3rd: .lff

    width = 5
    if ext == ".shp":
        width = 6
    if ext == ".lff":
        width = 7

    return FontFace(
        filename=font_path.name,  # "txt.shx", "simplex.shx", ...
        family=font_path.stem.lower(),  # "txt", "simplex", ...
        style=font_path.suffix.lower(),  # ".shx", ".shp" or ".lff"
        width=width,
        weight=400,
    )



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest

import ezdxf
from ezdxf.blkrefs import BlockReferenceCounter, find_unreferenced_blocks


def test_non_exiting_handles_return_0():
    doc = ezdxf.new()
    ref_counter = BlockReferenceCounter(doc)
    assert ref_counter.by_handle("xyz") == 0, "not existing handles should return 0"
    assert ref_counter.by_name("xyz") == 0, "not existing block name should return 0"


def test_access_interface():
    doc = ezdxf.new()
    msp = doc.modelspace()
    block = doc.blocks.new("First")
    msp.add_blockref("First", (0, 0))
    ref_counter = BlockReferenceCounter(doc)
    assert ref_counter.by_handle(block.block_record_handle) == 1
    assert ref_counter.by_name("First") == 1


def test_count_simple_references():
    count = 10
    doc = ezdxf.new()
    doc.blocks.new("First")
    msp = doc.modelspace()
    psp = doc.layout()
    for _ in range(count):
        msp.add_blockref("First", (0, 0))
        psp.add_blockref("First", (0, 0))
    ref_counter = BlockReferenceCounter(doc)
    assert ref_counter.by_name("First") == 20


def test_count_nested_block_references():
    count = 10
    doc = ezdxf.new()
    block1 = doc.blocks.new("First")
    block2 = doc.blocks.new("Second")
    block1.add_blockref(block2.dxf.name, (0, 0))

    msp = doc.modelspace()
    for _ in range(count):
        msp.add_blockref("First", (0, 0))
    ref_counter = BlockReferenceCounter(doc)
    assert ref_counter.by_name("First") == 10
    assert ref_counter.by_name("Second") == 1, "referenced only once in block First"


def test_count_references_used_in_xdata():
    doc = ezdxf.new()
    msp = doc.modelspace()
    block = doc.blocks.new("First")
    handle = block.block_record.dxf.handle
    line = msp.add_line((0, 0), (1, 0))

    # attach XDATA handle to block reference
    line.set_xdata("ezdxf", [(1005, handle)])
    ref_counter = BlockReferenceCounter(doc)
    assert ref_counter.by_handle(handle) == 1


def test_count_references_used_in_app_data():
    doc = ezdxf.new()
    msp = doc.modelspace()
    block = doc.blocks.new("First")
    handle = block.block_record.dxf.handle
    line = msp.add_line((0, 0), (1, 0))

    # attach XDATA handle to block reference
    line.set_app_data("ezdxf", [(320, handle), (480, handle)])
    ref_counter = BlockReferenceCounter(doc)
    assert ref_counter.by_handle(handle) == 2


def test_count_references_used_in_xrecord():
    doc = ezdxf.new()
    block = doc.blocks.new("First")
    handle = block.block_record.dxf.handle
    xrecord = doc.rootdict.add_xrecord("Test")
    xrecord.reset([(320, handle), (330, handle), (480, handle)])

    ref_counter = BlockReferenceCounter(doc)
    assert ref_counter.by_handle(handle) == 3


def test_count_references_in_header_section():
    doc = ezdxf.new()
    doc.blocks.new("Arrow")
    for var_name in ("$DIMBLK", "$DIMBLK1", "$DIMBLK2", "$DIMLDRBLK"):
        doc.header[var_name] = "Arrow"

    ref_counter = BlockReferenceCounter(doc)
    assert ref_counter.by_name("Arrow") == 4


def test_count_references_in_dimstyle():
    doc = ezdxf.new()
    doc.blocks.new("Arrow")
    dimstyle = doc.dimstyles.new("Test")
    dimstyle.dxf.dimblk = "Arrow"
    dimstyle.dxf.dimblk1 = "Arrow"
    dimstyle.dxf.dimblk2 = "Arrow"
    dimstyle.dxf.dimldrblk = "Arrow"

    ref_counter = BlockReferenceCounter(doc)
    assert ref_counter.by_name("Arrow") == 4


def test_count_references_in_leader():
    doc = ezdxf.new()
    doc.blocks.new("Arrow")
    msp = doc.modelspace()
    msp.add_leader(
        vertices=[(0, 0), (1, 0)],
        override={
            "dimblk": "Arrow",  # ignored by LEADER
            "dimblk1": "Arrow",  # ignored by LEADER
            "dimblk2": "Arrow",  # ignored by LEADER
            "dimldrblk": "Arrow",  # stored in XDATA
        },
    )

    ref_counter = BlockReferenceCounter(doc)
    assert ref_counter.by_name("Arrow") == 1


def test_count_references_for_anonymous_dimension_block():
    doc = ezdxf.new()
    msp = doc.modelspace()
    dim = msp.add_linear_dim(
        base=(25, 10),
        p1=(0, 0),
        p2=(50, 0),
    )
    dim.render()
    ref_counter = BlockReferenceCounter(doc)
    assert ref_counter.by_name(dim.dimension.dxf.geometry) == 1


def test_count_references_in_mleader_style():
    doc = ezdxf.new()
    arrow = doc.blocks.new("Arrow")
    arrow_handle = arrow.block_record.dxf.handle
    block = doc.blocks.new("Block")
    block_handle = block.block_record.dxf.handle
    mleader_style = doc.mleader_styles.new("Test")
    mleader_style.dxf.arrow_head_handle = arrow_handle
    mleader_style.dxf.block_record_handle = block_handle

    ref_counter = BlockReferenceCounter(doc)
    assert ref_counter.by_handle(arrow_handle) == 1
    assert ref_counter.by_handle(block_handle) == 1


class TestFindUnreferencedBlocks:
    def test_empty_document_has_no_unreferenced_blocks(self):
        doc = ezdxf.new()

        assert len(find_unreferenced_blocks(doc)) == 0

    def test_add_unreferenced_block(self):
        doc = ezdxf.new()
        doc.blocks.new("DUMMY")

        names = find_unreferenced_blocks(doc)
        assert len(names) == 1, "one unreferenced block expected"
        assert "DUMMY" in names, "name of unreferenced block expected"

    def test_add_referenced_block(self):
        doc = ezdxf.new()
        doc.blocks.new("DUMMY")
        msp = doc.modelspace()
        msp.add_blockref("DUMMY", (0, 0))

        names = find_unreferenced_blocks(doc)
        assert len(names) == 0, "no unreferenced block expected"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License


from ezdxf.lldxf.repair import filter_invalid_point_codes


def test_invalid_y_coord_after_xyz():
    result = list(
        filter_invalid_point_codes([(10, 1), (20, 2), (30, 3), (20, 0)])
    )
    assert result == [(10, 1), (20, 2), (30, 3)]


def test_invalid_y_coord_after_xy():
    result = list(filter_invalid_point_codes([(10, 1), (20, 2), (21, 2)]))
    assert result == [(10, 1), (20, 2)]


def test_z_axis_after_valid_point():
    result = list(
        filter_invalid_point_codes([(10, 1), (20, 2), (30, 3), (30, 0)])
    )
    assert result == [(10, 1), (20, 2), (30, 3)]


def test_misplaced_z_axis():
    result = list(
        filter_invalid_point_codes(
            [(10, 1), (20, 2), (1, "Text"), (30, 0), (1, "xxx")]
        )
    )
    assert result == [(10, 1), (20, 2), (1, "Text"), (1, "xxx")]


def test_correct_xy_axis():
    result = list(
        filter_invalid_point_codes([(10, 1), (20, 2), (10, 1), (20, 0)])
    )
    assert result == [(10, 1), (20, 2), (10, 1), (20, 0)]


def test_invalid_single_x_axis():
    result = list(filter_invalid_point_codes([(10, 1), (20, 2), (10, 1)]))
    assert result == [(10, 1), (20, 2)]


def test_preserve_leading_0_tag():
    result = list(
        filter_invalid_point_codes([(0, "SECTION"), (10, 1), (20, 2)])
    )
    assert result == [(0, "SECTION"), (10, 1), (20, 2)]


def test_preserve_elevation_group_code_38():
    result = list(filter_invalid_point_codes([(10, 1), (20, 2), (38, 0)]))
    assert result == [(10, 1), (20, 2), (38, 0)]



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License

import pytest
import struct
from ezdxf.tools.binarydata import ByteStream


def test_init():
    bs = ByteStream(b"ABCDABC\x00")
    assert bs.index == 0
    assert len(bs.buffer) == 8


def test_read_ps():
    bs = ByteStream(b"ABCDABC\x00")
    s = bs.read_padded_string()
    assert s == "ABCDABC"
    assert bs.index == 8
    assert bs.has_data is False


def test_read_ps_align():
    bs = ByteStream(b"ABCD\x00")
    s = bs.read_padded_string()
    assert s == "ABCD"
    assert bs.index == 8
    assert bs.has_data is False


def test_read_pus():
    bs = ByteStream(b"A\x00B\x00C\x00D\x00\x00\x00")
    s = bs.read_padded_unicode_string()
    assert s == "ABCD"
    assert bs.index == 12
    assert bs.has_data is False


def test_read_doubles():
    data = struct.pack("3d", 1.0, 2.0, 3.0)
    bs = ByteStream(data)
    x = bs.read_struct("d")[0]
    y = bs.read_struct("d")[0]
    z = bs.read_struct("d")[0]
    assert (x, y, z) == (1.0, 2.0, 3.0)
    assert bs.index == 24
    assert bs.has_data is False


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2011-2021, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.lldxf.const import DXF12, DXF2000, DXF2018, DXF2013, DXF2007, DXF2010
from ezdxf.lldxf.tags import Tags
from ezdxf.sections.header import HeaderSection
from ezdxf.lldxf.validator import header_validator
from ezdxf.sections.headervars import version_specific_group_code

INVALID_HEADER_STRUCTURE = """   9
$ACADVER
  1
AC1009
  100
$INSBASE
 10
0.0
 20
0.0
 30
0.0
"""

INVALID_HEADER_VAR_NAME = """   9
$ACADVER
  1
AC1009
  9
INSBASE
 10
0.0
 20
0.0
 30
0.0
"""

MINIMALISTIC_DXF12 = """  0
SECTION
  2
ENTITIES
  0
ENDSEC
  0
EOF
"""


def test_new_drawing():
    dwg = ezdxf.new("AC1009")
    assert "AC1009" == dwg.dxfversion


def test_valid_header():
    tags = Tags.from_text(TESTHEADER)[2:-1]
    result = list(header_validator(tags))
    assert 8 == len(result)


def test_invalid_header_structure():
    tags = Tags.from_text(INVALID_HEADER_STRUCTURE)
    with pytest.raises(ezdxf.DXFStructureError):
        list(header_validator(tags))


def test_invalid_header_var_name():
    tags = Tags.from_text(INVALID_HEADER_VAR_NAME)
    with pytest.raises(ezdxf.DXFValueError):
        list(header_validator(tags))


@pytest.fixture
def header():
    tags = Tags.from_text(TESTHEADER)
    tags.pop()  # remove 'ENDSEC'
    header = HeaderSection.load(tags)
    return header


def test_get_acadver(header):
    result = header["$ACADVER"]
    assert "AC1009" == result


def test_get_insbase(header):
    result = header["$INSBASE"]
    assert (0.0, 0.0, 0.0) == result


def test_getitem_keyerror(header):
    with pytest.raises(ezdxf.DXFKeyError):
        var = header["$TEST"]


def test_get(header):
    result = header.get("$TEST", "TEST")
    assert "TEST" == result


def test_set_existing_var(header):
    header["$ACADVER"] = "AC666"
    assert "AC666" == header["$ACADVER"]


def test_set_existing_point(header):
    header["$INSBASE"] = (1, 2, 3)
    assert (1, 2, 3) == header["$INSBASE"]


def test_set_unknown_var(header):
    with pytest.raises(ezdxf.DXFKeyError):
        header["$TEST"] = "test"


def test_create_var(header):
    header["$LIMMAX"] = (10, 20)
    assert (10, 20) == header["$LIMMAX"]


def test_create_var_wrong_args_2d(header):
    header["$LIMMAX"] = (10, 20, 30)
    assert (10, 20) == header["$LIMMAX"]


def test_create_var_wrong_args_3d(header):
    with pytest.raises(ezdxf.DXFValueError):
        header["$PUCSORG"] = (10, 20)


def test_contains(header):
    assert "$ACADVER" in header


def test_not_contains(header):
    assert "$MOZMAN" not in header


def test_remove_headervar(header):
    del header["$ACADVER"]
    assert "$ACADVER" not in header


def test_str_point(header):
    insbase_str = str(header.hdrvars["$INSBASE"])
    assert INSBASE == insbase_str


def test_new_dxf12():
    header = HeaderSection.new(ezdxf.const.DXF12)
    assert header["$ACADVER"] == DXF12


@pytest.fixture
def header_custom():
    tags = Tags.from_text(TESTCUSTOMPROPERTIES)
    tags.pop()  # remove 'ENDSEC'
    header = HeaderSection.load(tags)
    return header


def test_custom_properties_exists(header_custom):
    assert header_custom.custom_vars.has_tag("Custom Property 1")


def test_order_of_occurrence(header_custom):
    properties = header_custom.custom_vars.properties
    assert ("Custom Property 1", "Custom Value 1") == properties[0]
    assert ("Custom Property 2", "Custom Value 2") == properties[1]


def test_get_custom_property(header_custom):
    assert "Custom Value 1" == header_custom.custom_vars.get(
        "Custom Property 1"
    )


def test_get_custom_property_2(header_custom):
    assert "Custom Value 2" == header_custom.custom_vars.get(
        "Custom Property 2"
    )


def test_add_custom_property(header_custom):
    header_custom.custom_vars.append("Custom Property 3", "Custom Value 3")
    assert 3 == len(header_custom.custom_vars)
    assert "Custom Value 3" == header_custom.custom_vars.get(
        "Custom Property 3"
    )


def test_remove_custom_property(header_custom):
    header_custom.custom_vars.remove("Custom Property 1")
    assert 1 == len(header_custom.custom_vars)


def test_remove_not_existing_property(header_custom):
    with pytest.raises(ValueError):
        header_custom.custom_vars.remove("Does not Exist")


def test_replace_custom_property(header_custom):
    header_custom.custom_vars.replace("Custom Property 1", "new value")
    assert "new value" == header_custom.custom_vars.get("Custom Property 1")


def test_replace_not_existing_property(header_custom):
    with pytest.raises(ValueError):
        header_custom.custom_vars.replace("Does not Exist", "new value")


@pytest.mark.parametrize(
    "name,version,expected",
    [
        ("$ACADMAINTVER", DXF2018, 90),
        ("$ACADMAINTVER", DXF2013, 70),
        ("$XCLIPFRAME", DXF2007, 290),
        ("$XCLIPFRAME", DXF2010, 280),
    ],
)
def test_version_specific_header_vars(name, version, expected):
    assert version_specific_group_code(name, version) == expected


def test_version_specific_group_code_raises_key_error_for_unknown_names():
    with pytest.raises(KeyError):
        version_specific_group_code("$MOZMAN", DXF2000)


INSBASE = """ 10
0.0
 20
0.0
 30
0.0
"""

TESTHEADER = """  0
SECTION
  2
HEADER
  9
$ACADVER
  1
AC1009
  9
$INSBASE
 10
0.0
 20
0.0
 30
0.0
  9
$EXTMIN
 10
1.0000000000000000E+020
 20
1.0000000000000000E+020
 30
1.0000000000000000E+020
  9
$EXTMAX
 10
-1.0000000000000000E+020
 20
-1.0000000000000000E+020
 30
-1.0000000000000000E+020
  0
ENDSEC
"""

TESTCUSTOMPROPERTIES = """  0
SECTION
  2
HEADER
  9
$ACADVER
  1
AC1009
  9
$CUSTOMPROPERTYTAG
  1
Custom Property 1
  9
$CUSTOMPROPERTY
  1
Custom Value 1
  9
$CUSTOMPROPERTYTAG
  1
Custom Property 2
  9
$CUSTOMPROPERTY
  1
Custom Value 2
  0
ENDSEC
"""
